#include "ActionLookupTable.h"
#include "Log.h"
#include "State.h"

#include <cfloat>


ActionLookupTable::ActionLookupTable( State* aStatePtr )
  : ActionValueFunction( aStatePtr ),
    LookupTable( aStatePtr )
{
  // den Funktionspointer initialisieren
  value = 0.0;
  initWeightsFunctionPtr = &getValue;
}


ActionLookupTable::ActionLookupTable( const std::string& fileName )
  : ActionValueFunction( 0 ),
    LookupTable( 0 )
{
  // den Funktionspointer initialisieren
  value = 0.0;
  initWeightsFunctionPtr = &getValue;

  // Laden
  ActionValueFunction::load( fileName );
}


ActionLookupTable::~ActionLookupTable()
{
}


void ActionLookupTable::initWeights( const double (*initFunctionPtr)() )
{
  LookupTable::initWeights( initFunctionPtr );
}


void ActionLookupTable::initEligibilityTraces()
{
  LookupTable::initEligibilityTraces();
}


void ActionLookupTable::updateWeights( const double ALPHA, const double DELTA )
{
  LookupTable::updateWeights( ALPHA, DELTA );
}


void ActionLookupTable::updateEligibilityTraces( const double GAMMA, const double LAMBDA )
{
  LookupTable::updateEligibilityTraces( GAMMA, LAMBDA );
}


void ActionLookupTable::create()
{
  if ( ACTIONS > 0 ) {
    // Speicher holen
    LookupTable::allocateMemory();

    // indexOffset
    for ( unsigned int v = 0; v < ActionValueFunction::statePtr->getNumberOfVariables(); ++v ) {
      indexOffset[v] = 1;
      for ( unsigned int i = v + 1; i < ActionValueFunction::statePtr->getNumberOfVariables(); ++i ) {
        // +1 --> damit lower und upper Bound im Bereich sind
        indexOffset[v] *= (ActionValueFunction::statePtr->getMaxValueInt( i ) - ActionValueFunction::statePtr->getMinValueInt( i ) + 1);
      }
      indexOffset[v] *= ACTIONS;
    }

    // Anzahl der Gewichte bestimmen
    ARRAY_SIZE = ACTIONS;
    for ( unsigned int v = 0; v < ActionValueFunction::statePtr->getNumberOfVariables(); ++v ) {
      // +1 --> damit lower und upper Bound im Bereich sind
      ARRAY_SIZE *= (ActionValueFunction::statePtr->getMaxValueInt( v ) - ActionValueFunction::statePtr->getMinValueInt( v ) + 1);
    }

    // Gewichtsvektor anlegen
    weights = allocateMemoryWeights();

    // Gewichtsvektor initialisieren
    initWeights( initWeightsFunctionPtr );
  }
  else {
    RL_LOGERROR( "No actions registered! --> Register actions first!" );
  }
}


void ActionLookupTable::accumulatingTraces( const double VALUE )
{
  if ( traceStorage == Types::ARRAY ) {
    if ( eTraces == 0 ) {
      RL_LOGERROR( "traceStorage == ARRAY and eTraces == 0"
                   << " --> Set traceStorage to MAP or call initEligibilityTraces before calling updateWeights!" );
      return;
    }

    if ( activeIndex != -1 ) {
      eTraces[ activeIndex + activeActionIndex ] += VALUE;
    }
  }
  else if ( traceStorage == Types::MAP ) {
    std::map< int, double >::iterator i;

    if ( activeIndex != -1 ) {
      i = eTracesMap.find( activeIndex + activeActionIndex );
      if ( i == eTracesMap.end() ) {
        // Key existiert noch nicht --> anlegen
        eTracesMap[ activeIndex + activeActionIndex ] = VALUE;
      }
      else {
        // Key existiert schon --> Wert aufaddieren
        i->second += VALUE;
      }
    }
  }
  else {
    RL_LOGERROR( "Unknown traceStorage = " << traceStorage );
  }
}


void ActionLookupTable::replacingTraces( const double VALUE )
{
  if ( traceStorage == Types::ARRAY ) {
    if ( eTraces == 0 ) {
      RL_LOGERROR( "traceStorage == ARRAY and eTraces == 0"
                   << " --> Set traceStorage to MAP or call initEligibilityTraces before calling updateWeights!" );
      return;
    }

    if ( activeIndex != -1 ) {
      for ( int a = 0; a < ACTIONS; ++a ) {
        eTraces[ activeIndex + activeActionIndex ] = ((a == activeActionIndex) ? VALUE : 0.0);
      }
    }
  }
  else if ( traceStorage == Types::MAP ) {
    std::map< int, double >::iterator i;
    int index = 0;

    if ( activeIndex != -1 ) {
      for ( int a = 0; a < ACTIONS; ++a ) {
        index = activeIndex + a;
        if ( a == activeActionIndex ) {
          // neuen Wert setzen
          eTracesMap[ index ] = VALUE;
        }
        else {
          // falls der Key existiert --> loeschen
          i = eTracesMap.find( index );
          if ( i != eTracesMap.end() ) {
            // Key loeschen
            eTracesMap.erase( i );
          }
        }
      }
    }
  }
  else {
    RL_LOGERROR( "Unknown traceStorage = " << traceStorage );
  }
}


void ActionLookupTable::storeActiveIndex( const int actionIndex )
{
  if ( activeIndex != -1 ) {
    storedActiveIndex = activeIndex + actionIndex;
  }
  else {
    storedActiveIndex = -1;
  }
}


const double ActionLookupTable::getActionValue( const int actionIndex ) const
{
  if ( weights != 0 ) {
    if ( (actionIndex < 0) || (actionIndex >= ACTIONS) ) {
      RL_LOGERROR( "actionIndex[" << actionIndex<< "] out of bounds [ 0, " << ACTIONS << ")" );
    }

    if ( activeIndex != -1 ) {
      return weights[ activeIndex + actionIndex ];
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }
  return DBL_MAX;
}


void ActionLookupTable::storeEpisodeState( const int actionIndex )
{
  if ( activeIndex != -1 ) {
    episodeIndexHistorySet.insert( activeIndex + actionIndex );
  }
}


void ActionLookupTable::setActionValue( const int actionIndex, const double value )
{
  if ( weights != 0 ) {
    if ( (actionIndex < 0) || (actionIndex >= ACTIONS) ) {
      RL_LOGERROR( "actionIndex[" << actionIndex<< "] out of bounds [ 0, " << ACTIONS << ")" );
    }

    for ( int i = actionIndex; i < ARRAY_SIZE; i += ACTIONS ) {
      weights[i] = value;
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }
}


void ActionLookupTable::setActiveActionValue( const int actionIndex, const double value )
{
  if ( weights != 0 ) {
    if ( (actionIndex < 0) || (actionIndex >= ACTIONS) ) {
      RL_LOGERROR( "actionIndex[" << actionIndex<< "] out of bounds [ 0, " << ACTIONS << ")" );
    }

    if ( activeIndex != -1 ) {
      weights[ activeIndex + actionIndex ] = value;
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }
}


void ActionLookupTable::storeState()
{
  storeActiveIndex( activeActionIndex );
  storeEpisodeState( activeActionIndex );
}


void ActionLookupTable::setState()
{
  LookupTable::setState();
}


void ActionLookupTable::initEpisodeStateHistory()
{
  LookupTable::initEpisodeStateHistory();
}


void ActionLookupTable::valueEpisode( const double value )
{
  LookupTable::valueEpisode( value );
}


void ActionLookupTable::checkWeightLimit()
{
  LookupTable::checkWeightLimit( minWeightLimit, maxWeightLimit );
}


const double ActionLookupTable::getMinWeight() const
{
  return LookupTable::getMinWeight();
}


const double ActionLookupTable::getAvgWeight( const bool abs ) const
{
  return LookupTable::getAvgWeight( abs );
}


const double ActionLookupTable::getMaxWeight() const
{
  return LookupTable::getMaxWeight();
}


void ActionLookupTable::removeAction( const int actionIndex )
{
  if ( ACTIONS == 1 ) {
    RL_LOGERROR( "ACTIONS == 1 --> Can not remove last action!" );
    return;
  }
  if ( weights == 0 ) {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
    return;
  }
  if ( (actionIndex < 0) || (actionIndex >= ACTIONS) ) {
    RL_LOGERROR( "actionIndex[" << actionIndex<< "] out of bounds [ 0, " << ACTIONS << ")" );
    return;
  }

  // neue Anzahl von Aktionen
  int NEW_ACTIONS = ACTIONS - 1;

  // indexOffset
  for ( unsigned int d = 0; d < ActionValueFunction::statePtr->getNumberOfVariables(); ++d ) {
    indexOffset[d] = 0;
    for ( unsigned int i = d + 1; i < ActionValueFunction::statePtr->getNumberOfVariables(); ++i ) {
      indexOffset[d] *= (ActionValueFunction::statePtr->getMaxValueInt( d ) - ActionValueFunction::statePtr->getMinValueInt( d ));
    }
    indexOffset[d] *= NEW_ACTIONS;
  }

  // Weight anlegen
  int NEW_ARRAY_SIZE = 0;
  for ( unsigned int d = 0; d < ActionValueFunction::statePtr->getNumberOfVariables(); ++d ) {
    NEW_ARRAY_SIZE *= (ActionValueFunction::statePtr->getMaxValueInt( d ) - ActionValueFunction::statePtr->getMinValueInt( d ));
  }
  NEW_ARRAY_SIZE *= NEW_ACTIONS;
  // Speicher fuer weights holen
  double* newWeights = new double[ NEW_ARRAY_SIZE ];

  // alle Werte bis auf actionIndex in neues Array kopieren
  int weightIndex = ((actionIndex > 0) ? 0 : 1);
  int newWeightIndex = 0;
  int size = ((actionIndex > 0) ? actionIndex : NEW_ACTIONS);

  while ( weightIndex < ARRAY_SIZE ) {
    memcpy( &newWeights[newWeightIndex], &weights[weightIndex], size * sizeof( double ) );
    weightIndex += size + 1;
    newWeightIndex += size;
    size = std::min( NEW_ACTIONS, ARRAY_SIZE - weightIndex );
  }

  // ab jetzt eine Aktion weniger moeglich
  ACTIONS = NEW_ACTIONS;
  // neue Arraygroesse merken
  ARRAY_SIZE = NEW_ARRAY_SIZE;
  // Gewichtsvektor ist nun reduziert --> alten Speicher freigeben
  delete[] weights;
  weights = newWeights;

  // traces freigeben
  if ( (traceStorage == Types::ARRAY) && ( eTraces != 0 ) ) {
    delete[] eTraces;
  }
  else if ( (traceStorage == Types::MAP) && (eTracesMap.empty() == false) ) {
    eTracesMap.clear();
  }

  // activeIndex ungueltig
  activeIndex = -1;
  storedActiveIndex = -1;

  // alle gespeicherten Indizees sind ungueltig
  if ( episodeIndexHistorySet.empty() == false ) {
    episodeIndexHistorySet.clear();
  }
}


const int ActionLookupTable::save( FILE* file ) const
{
  return LookupTable::save( file );
}


const int ActionLookupTable::load( FILE* file )
{
  // jetzt ist der StatePtr bereits geladen --> auch im LookupTable setzen
  LookupTable::setStatePtr( ValueFunction::statePtr );
  return LookupTable::load( file );
}
