#ifndef __ACTION_NEURAL_NETWORK_H__
#define __ACTION_NEURAL_NETWORK_H__

#include "ActionValueFunction.h"
#include "NeuralNetwork.h"

#include <list>
#include <string>

class State;


class ActionNeuralNetwork : public ActionValueFunction,
                            public NeuralNetwork
{
  public:
    /**
    * Konstruktor
    * @param aStatePtr - Pointer auf den Zustand
    * @param numberOfHiddenNeurons - Anzahl der Neuronen im Hiddenlayer
    */
    ActionNeuralNetwork( State* aStatePtr, const int numberOfHiddenNeurons );

    /**
    * Konstruktor zum Laden
    * @param fileName - Name der zu ladenden Datei
    */
    ActionNeuralNetwork( const std::string& fileName );

    /** Destruktor */
    virtual ~ActionNeuralNetwork();

    /**
    * Initialisiert die Gewichte
    * @param initFunctionPtr - Pointer auf die Funktion die zum Initialisieren benutzt wird
    */
    virtual void initWeights( const double (*initFunctionPtr)() );

    /**
    * Initialisieren der EligibilityTraces mit 0.0
    */
    virtual void initEligibilityTraces();

    /**
    * Updaten der Gewichte
    * @param ALPHA - die Lernrate
    * @param DELTA - die Aenderung des Gewichts
    */
    virtual void updateWeights( const double ALPHA, const double DELTA );

    /**
    * Updaten der EligibilityTraces
    * @param GAMMA - Discountrate
    * @param LAMBDA - TraceDecay
    */
    virtual void updateEligibilityTraces( const double GAMMA, const double LAMBDA );

    /**
    * Erzeugt den Gewichtsvektor --> darf erst gerufen werden nachdem alle Actions registriert wurden
    */
    virtual void create();

    /**
    * In dieser Methode passiert nichts --> wird alles in updateEligibilityTraces erledigt
    * @param VALUE - der Wert um den erhoeht wird
    */
    virtual void accumulatingTraces( const double VALUE );

    /**
    * In dieser Methode passiert nichts --> wird alles in updateEligibilityTraces erledigt
    * @param VALUE - der Wert auf den gesetzt wird
    */
    virtual void replacingTraces( const double VALUE );

    /**
    * Liefert den Wert einer Aktion in einem Zustands mit dem zuvor aktivierten Index.
    *@param actionIndex - der Index der Aktion
    * @return Wert der Aktion
    */
    virtual const double getActionValue( const int actionIndex ) const;

    /**
    * Ueberprueft ob die Gewichte in den Grenzen sind.
    * minWeight <= weight <= maxWeight
    */
    virtual void checkWeightLimit();

    /**
    * Liefert das minimale Gewicht zurueck.
    * @return das minimale Gewicht
    */
    virtual const double getMinWeight() const;

    /**
    * Liefert das durchschnittliche Gewicht zurueck.
    * @param abs - wenn true wird der Durchschnitt ueber den absoluten Betrag der Gewichte gebildet
    * @return das durchschnittliche Gewicht
    */
    virtual const double getAvgWeight( const bool abs ) const;

    /**
    * Liefert das maximale Gewicht zurueck.
    * @return das maximale Gewicht
    */
    virtual const double getMaxWeight() const;

    /**
    * Setzt den Zustand
    */
    virtual void setState();

    /**
    * Falls keine EligibilityTraces verwendet werden, muss der Zustand gespeichert
    * werden, damit man weiss welche Gewichte geupdatet werden muessen.
    * Beim Neuronalen Netz entspricht dies aber updateEligibilityTraces
    * mit GAMMA = LAMBDA = 1
    */
    virtual void storeState();

    /**
    * Initialisiert die episodeStateHistory.
    * Nicht implementiert, da ich nicht sicher bin, ob das Sinn macht.
    */
    virtual void initEpisodeStateHistory();

    /**
    * Bewertet die gesamte Episode.
    * --> dafuer muss episodeStateHistory auf true gesetzt sein
    * Nicht implementiert, da ich nicht sicher bin, ob das Sinn macht.
    * @param value - dieser Wert wird auf alle Gewichte der Episode gerechnet
    */
    virtual void valueEpisode( const double value );


  protected:
    /**
    * Entfernt die Aktion aus dem Neural Network und reduziert die Gesamtanzahl der Aktionen.
    * Achtung! EligibilityTraces werden geloescht, es gibt keinen aktiven Index mehr und die Indexhistory wird geleert!
    * @param der Index der Aktion die entfernt werden soll
    */
    virtual void removeAction( const int actionIndex );
    
    /**
    * Abspeichern
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int save( FILE* file ) const;

    /**
    * Laden
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int load( FILE* file );
};

#endif
