#ifndef __ACTION_TILE_CODING_H__
#define __ACTION_TILE_CODING_H__

#include "ActionValueFunction.h"
#include "TileCoding.h"

class ActionTilings;


class ActionTileCoding : public ActionValueFunction,
                         public TileCoding
{
  public:
    /**
    * Konstruktor
    * @param aStatePtr - Pointer auf den Zustand
    */
    ActionTileCoding( State* aStatePtr );

    /**
    * Konstruktor zum Laden
    * @param fileName - Name der zu ladenden Datei
    */
    ActionTileCoding( const std::string& fileName );

    /**
    * Destruktor
    */
    ~ActionTileCoding();

    /**
    * Initialisiert die Gewichte
    * --> Diese Implementation ruft nur die Implementation aus TileCoding auf!
    * @param initFunctionPtr - Pointer auf die Funktion die zum Initialisieren bentzt wird
    */
    virtual void initWeights( const double (*initFunctionPtr)() );

    /**
    * Initialisieren der EligibilityTraces mit 0.0
    * --> Diese Implementation ruft nur die Implementation aus TileCoding auf!
    */
    virtual void initEligibilityTraces();

    /**
    * Updaten der Gewichte
    * --> Diese Implementation ruft nur die Implementation aus TileCoding auf!
    * @param ALPHA - die Lernrate
    * @param DELTA - die Aenderung des Gewichts
    */
    virtual void updateWeights( const double ALPHA, const double DELTA );

    /**
    * Updaten der EligibilityTraces
    * --> Diese Implementation ruft nur die Implementation aus TileCoding auf!
    * @param GAMMA - Discountrate
    * @param LAMBDA - TraceDecay
    */
    virtual void updateEligibilityTraces( const double GAMMA, const double LAMBDA );

    /**
    * Anmelden eines Tilings
    * @param actionTilingPtr - Pointer auf ein Action-Tilingsobjekt
    */
    virtual void registerTiling( ActionTilings* actionTilingPtr );

    /**
    * Liefert den Wert einer Aktion fuer den aktuellen Zustand.
    * @param der Index der Aktion
    * @return der Wert der Aktion
    */
    virtual const double getActionValue( const int actionIndex ) const;

    /**
    * Setzt den Wert fuer eine Aktion. 
    * Sinn: Soll optimistische/pessimistische Werte fuer einzelne Aktionen ermoeglichen,
            damit diese einen Vorteil/Nachteil erhalten.
      @param actionIndex - der Index der Aktion
      @param value - der Wert der Aktion
    */
    virtual void setActionValue( const int actionIndex, const double value );

    /**
    * Die Traces der aktivierten Features werden erhoeht.
    * @param VALUE - der Wert um den erhoeht wird
    */
    virtual void accumulatingTraces( const double VALUE );

    /**
    * Die Traces der aktivierten Features werden gesetzt.
    * @param VALUE - der Wert auf den gesetzt wird
    */
    virtual void replacingTraces( const double VALUE );

    /**
    * Falls keine EligibilityTraces verwendet werden, muss der Zustand gespeichert
    * werden, damit man weiss welche Gewichte geupdatet werden muessen.
    * Beim TileCoding werden dazu die aktivierten Tiles gespeichert.
    * --> Diese Implementation ruft nur die Implementation aus TileCoding auf!
    */
    virtual void storeState();

    /**
    * Ueberprueft ob die Gewichte in den Grenzen sind.
    * minWeight <= weight <= maxWeight
    */
    virtual void checkWeightLimit();

    /**
    * Liefert das minimale Gewicht zurueck.
    * --> Diese Implementation ruft nur die Implementation aus TileCoding auf!
    * @return das minimale Gewicht
    */
    virtual const double getMinWeight() const;

    /**
    * Liefert das durchschnittliche Gewicht zurueck.
    * --> Diese Implementation ruft nur die Implementation aus TileCoding auf!
    * @param abs - wenn true wird der Durchschnitt ueber den absoluten Betrag der Gewichte gebildet
    * @return das durchschnittliche Gewicht
    */
    virtual const double getAvgWeight( const bool abs ) const;

    /**
    * Liefert das maximale Gewicht zurueck.
    * --> Diese Implementation ruft nur die Implementation aus TileCoding auf!
    * @return das maximale Gewicht
    */
    virtual const double getMaxWeight() const;

    /**
    * Setzt den Zustand --> aktiviert die Tilings
    * --> Diese Implementation ruft nur die Implementation aus TileCoding auf!
    */
    virtual void setState();

    /**
    * Initialisiert die episodeStateHistory.
    * --> Diese Implementation ruft nur die Implementation aus TileCoding auf!
    */
    virtual void initEpisodeStateHistory();

    /**
    * Bewertet die gesamte Episode.
    * --> dafuer muss episodeStateHistory auf true gesetzt sein
    * --> Diese Implementation ruft nur die Implementation aus TileCoding auf!
    * @param value - dieser Wert wird auf alle Gewichte der Episode gerechnet
    */
    virtual void valueEpisode( const double value );


  protected:
    /**
    * Entfernt die Aktion aus dem ActionTileCoding und reduziert die Gesamtanzahl der Aktionen.
    * Achtung! EligibilityTraces werden geloescht, es gibt keine aktiven Tiles mehr und die Indexhistory wird geleert!
    * @param der Index der Aktion die entfernt werden soll
    */
    virtual void removeAction( const int actionIndex );

    /**
    * Abspeichern
    * --> Diese Implementation ruft nur die Implementation aus TileCoding auf!
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int save( FILE* file ) const;

    /**
    * Laden
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int load( FILE* file );
};

#endif
