#include "ActionTilings.h"
#include "Log.h"
#include "State.h"

#include <cmath>
#include <cfloat>



ActionTilings::ActionTilings( const int numberOfTilings, const std::list< Types::TILINGS_VAR_STRUCT >& varList )
  : Tilings( numberOfTilings, varList )
{
  // Anzahl der Aktionen --> wird spaeter bei create gesetzt
  ACTIONS = 0;
}


ActionTilings::ActionTilings( State* aStatePtr, const int numberOfActions, FILE* file )
  : Tilings( aStatePtr, file )
{
  // Anzahl der Aktionen
  ACTIONS = numberOfActions;
}


ActionTilings::~ActionTilings()
{
}


void ActionTilings::create( State* aStatePtr, const int numberOfActions )
{
  // Pointer auf den Zustand
  statePtr = aStatePtr;
  // Anzahl der Aktionen
  ACTIONS = numberOfActions;
  if ( ACTIONS <= 0 ) {
    RL_LOGERROR( "ACTIONS[" << ACTIONS << "] <= 0" );
  }

  // tiles
  for ( int d = 0; d < DIMENSIONS; ++d ) {
    tiles[d] = static_cast< int >( ceil( (statePtr->getMaxValueDbl( vars[d].id ) - statePtr->getMinValueDbl( vars[d].id )) / (vars[d].resolution * TILINGS) ) );
    if ( vars[d].tileType == Types::WRAP ) {
      tiles[d] += 1;
    }
  }

  // resolution neu setzen --> da nun die tiles ermittelt sind
  for ( int d = 0; d < DIMENSIONS; ++d ) {
    vars[d].resolution = (statePtr->getMaxValueDbl( vars[d].id ) - statePtr->getMinValueDbl( vars[d].id )) / ((tiles[d] - ((vars[d].tileType == Types::WRAP) ? 1 : 0)) * TILINGS);
  }

  // tileWidth
  for ( int d = 0; d < DIMENSIONS; ++d ) {
    tileWidth[d] = (statePtr->getMaxValueDbl( vars[d].id ) - statePtr->getMinValueDbl( vars[d].id )) / (tiles[d] - ((vars[d].tileType == Types::WRAP) ? 1 : 0));
  }

  // indexOffset
  for ( int d = 0; d < DIMENSIONS; ++d ) {
    indexOffset[d] = TILINGS;
    for ( int i = d + 1; i < DIMENSIONS; ++i ) {
      indexOffset[d] *= tiles[i];
    }
    indexOffset[d] *= ACTIONS;
  }

  // tileOffset
  for ( int t = 0; t < TILINGS; ++t ) {
    for ( int d = 0; d < DIMENSIONS; ++d ) {
      tileOffset[t][d] = vars[d].resolution * t;
    }
  }

  // Weight anlegen
  ARRAY_SIZE = TILINGS;
  for ( int d = 0; d < DIMENSIONS; ++d ) {
    ARRAY_SIZE *= tiles[d];
  }
  ARRAY_SIZE *= ACTIONS;
  // Speicher fuer weights holen
  weights = allocateMemoryWeights();
}


void ActionTilings::accumulatingTraces( const double VALUE, const int actionIndex )
{
  if ( traceStorage == Types::ARRAY ) {
    if ( eTraces == 0 ) {
      RL_LOGERROR( "traceStorage == ARRAY and eTraces == 0"
                   << " --> Set traceStorage to MAP or call initEligibilityTraces before calling updateWeights!" );
      return;
    }

    for ( int t = 0; t < TILINGS; ++t ) {
      if ( activeTiles[t] != -1 ) {
        eTraces[ activeTiles[t] + actionIndex * TILINGS ] += VALUE;
      }
    }
  }
  else if ( traceStorage == Types::MAP ) {
    std::map< int, double >::iterator i;
    int index;

    for ( int t = 0; t < TILINGS; ++t ) {
      if ( activeTiles[t] != -1 ) {
        index = activeTiles[t] + actionIndex * TILINGS;
        i = eTracesMap.find( index );
        if ( i == eTracesMap.end() ) {
          // Key existiert noch nicht --> anlegen
          eTracesMap[ index ] = VALUE;
        }
        else {
          // Key existiert schon --> Wert aufaddieren
          i->second += VALUE;
        }
      }
    }
  }
  else {
    RL_LOGERROR( "Unknown traceStorage = " << traceStorage );
  }
}


void ActionTilings::replacingTraces( const double VALUE, const int actionIndex )
{
  if ( traceStorage == Types::ARRAY ) {
    if ( eTraces == 0 ) {
      RL_LOGERROR( "traceStorage == ARRAY and eTraces == 0"
                   << " --> Set traceStorage to MAP or call initEligibilityTraces before calling updateWeights!" );
      return;
    }

    for ( int t = 0; t < TILINGS; ++t ) {
      if ( activeTiles[t] != -1 ) {
        for ( int a = 0; a < ACTIONS; ++a ) {
          eTraces[ activeTiles[t] + a * TILINGS ] = ((a == actionIndex) ? VALUE : 0.0);
        }
      }
    }
  }
  else if ( traceStorage == Types::MAP ) {
    std::map< int, double >::iterator i;
    int index;

    for ( int t = 0; t < TILINGS; ++t ) {
      if ( activeTiles[t] != -1 ) {
        for ( int a = 0; a < ACTIONS; ++a ) {
          index = activeTiles[t] + a * TILINGS;
          if ( a == actionIndex ) {
            // neuen Wert setzen
            eTracesMap[ index ] = VALUE;
          }
          else {
            // falls der Key existiert --> loeschen
            i = eTracesMap.find( index );
            if ( i != eTracesMap.end() ) {
              // Key loeschen
              eTracesMap.erase( i );
            }
          }
        }
      }
    }
  }
  else {
    RL_LOGERROR( "Unknown traceStorage = " << traceStorage );
  }
}


void ActionTilings::storeActiveTiles( const int actionIndex )
{
  if ( storedActiveIndices == 0 ) {
    storedActiveIndices = new int[ TILINGS ];
  }

  for ( int t = 0; t < TILINGS; ++t ) {
    if ( activeTiles[t] != -1 ) {
      storedActiveIndices[t] = activeTiles[t] + actionIndex * TILINGS;
    }
    else {
      storedActiveIndices[t] = -1;
    }
  }
}


const double ActionTilings::getActionValue( const int actionIndex ) const
{
  if ( weights != 0 ) {
    if ( (actionIndex < 0) || (actionIndex >= ACTIONS) ) {
      RL_LOGERROR( "actionIndex[" << actionIndex<< "] out of bounds [ 0, " << ACTIONS << ")" );
    }

    double value = 0.0;
    for ( int t = 0; t < TILINGS; ++t ) {
      if ( activeTiles[t] != -1 ) {
        value += weights[ activeTiles[t] + actionIndex * TILINGS ];
      }
    }
    return value;
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
    return DBL_MAX;
  }
}


void ActionTilings::storeEpisodeState( const int actionIndex )
{
  for ( int t = 0; t < TILINGS; ++t ) {
    if ( activeTiles[t] != -1 ) {
      episodeIndexHistorySet.insert( activeTiles[t] + actionIndex * TILINGS );
    }
  }
}


void ActionTilings::setActionValue( const int actionIndex, const double value )
{
  if ( weights != 0 ) {
    if ( (actionIndex < 0) || (actionIndex >= ACTIONS) ) {
      RL_LOGERROR( "actionIndex[" << actionIndex<< "] out of bounds [ 0, " << ACTIONS << ")" );
    }

    for ( int i = (actionIndex * TILINGS); i < ARRAY_SIZE; i += (ACTIONS * TILINGS) ) {
      for ( int t = 0; t < TILINGS; ++t ) {
        weights[i+t] = value;
      }
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }
}


void ActionTilings::setActiveActionValue( const int actionIndex, const double value )
{
  if ( weights != 0 ) {
    if ( (actionIndex < 0) || (actionIndex >= ACTIONS) ) {
      RL_LOGERROR( "actionIndex[" << actionIndex<< "] out of bounds [ 0, " << ACTIONS << ")" );
    }

    for ( int t = 0; t < TILINGS; ++t ) {
      if ( activeTiles[t] != -1 ) {
        weights[ activeTiles[t] + actionIndex * TILINGS ] = value;
      }
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }
}


const bool ActionTilings::removeAction( const int actionIndex )
{
  if ( ACTIONS == 1 ) {
    RL_LOGERROR( "ACTIONS == 1 --> Can not remove last action!" );
    return false;
  }
  if ( weights == 0 ) {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
    return false;
  }
  if ( (actionIndex < 0) || (actionIndex >= ACTIONS) ) {
    RL_LOGERROR( "actionIndex[" << actionIndex<< "] out of bounds [ 0, " << ACTIONS << ")" );
    return false;
  }

  // neue Anzahl von Aktionen
  int NEW_ACTIONS = ACTIONS - 1;

  // indexOffset
  for ( int d = 0; d < DIMENSIONS; ++d ) {
    indexOffset[d] = TILINGS;
    for ( int i = d + 1; i < DIMENSIONS; ++i ) {
      indexOffset[d] *= tiles[i];
    }
    indexOffset[d] *= NEW_ACTIONS;
  }

  // tileOffset
  for ( int t = 0; t < TILINGS; ++t ) {
    for ( int d = 0; d < DIMENSIONS; ++d ) {
      tileOffset[t][d] = vars[d].resolution * t;
    }
  }

  // Weight anlegen
  int NEW_ARRAY_SIZE = TILINGS;
  for ( int d = 0; d < DIMENSIONS; ++d ) {
    NEW_ARRAY_SIZE *= tiles[d];
  }
  NEW_ARRAY_SIZE *= NEW_ACTIONS;
  // Speicher fuer weights holen
  double* newWeights = new double[ NEW_ARRAY_SIZE ];

  // alle Werte bis auf actionIndex in neues Array kopieren
  int weightIndex = ((actionIndex > 0) ? 0 : TILINGS);
  int newWeightIndex = 0;
  int size = ((actionIndex > 0) ? (actionIndex * TILINGS) : (NEW_ACTIONS * TILINGS));

  while ( weightIndex < ARRAY_SIZE ) {
    memcpy( &newWeights[newWeightIndex], &weights[weightIndex], size * sizeof( double ) );
    weightIndex += size + TILINGS;
    newWeightIndex += size;
    size = std::min( NEW_ACTIONS * TILINGS, ARRAY_SIZE - weightIndex );
  }

  // ab jetzt eine Aktion weniger moeglich
  ACTIONS = NEW_ACTIONS;
  // neue Arraygroesse merken
  ARRAY_SIZE = NEW_ARRAY_SIZE;
  // Gewichtsvektor ist nun reduziert --> alten Speicher freigeben
  delete[] weights;
  weights = newWeights;

  // traces freigeben
  if ( (traceStorage == Types::ARRAY) && ( eTraces != 0 ) ) {
    delete[] eTraces;
  }
  else if ( (traceStorage == Types::MAP) && (eTracesMap.empty() == false) ) {
    eTracesMap.clear();
  }
  else if ( storedActiveIndices != 0 ) {
    delete[] storedActiveIndices;
  }

  // activeTiles ungueltig
  for ( int t = 0; t < TILINGS; ++t ) {
    activeTiles[t] = -1;
  }

  // alle gespeicherten Indizees sind ungueltig
  if ( episodeIndexHistorySet.empty() == false ) {
    episodeIndexHistorySet.clear();
  }

  return true;
}
