#include "ActionValueFunction.h"
#include "AbstractAction.h"
#include "State.h"
#include "Utility.h"

#include <cerrno>
#include <cfloat>
#include <cmath>
#include <vector>



ActionValueFunction::ActionValueFunction( State* aStatePtr )
  : ValueFunction( aStatePtr )
{
  ACTIONS = 0;
  activeActionIndex = -1;
  actionValues = 0;
}


ActionValueFunction::~ActionValueFunction()
{
  if ( actionValues != 0 ) {
    delete[] actionValues;
  }
  for ( std::vector< const AbstractAction* >::iterator a = registeredActions.begin();
        a != registeredActions.end(); ++a )
  {
    delete *a;
  }
}


void ActionValueFunction::registerAction( const AbstractAction* actionPtr )
{
  registeredActions.push_back( actionPtr );
  ++ACTIONS;
}


void ActionValueFunction::unregisterAction( const AbstractAction* actionPtr )
{
  int actionIndex = 0;
  for ( std::vector< const AbstractAction* >::iterator a = registeredActions.begin();
        a != registeredActions.end(); ++a )
  {
    if ( *a == actionPtr ) {
      registeredActions.erase( a );
      removeAction( actionIndex );
      --ACTIONS;
      break;
    }
    ++actionIndex;
  }
}


void ActionValueFunction::unregisterAction( const int actionIndex )
{
  if ( (actionIndex < 0) || (actionIndex >= ACTIONS) ) {
    RL_LOGERROR( "actionIndex " << actionIndex << " out of bounds [0, " << ACTIONS << ")" );
  }
  else {
    registeredActions.erase( registeredActions.begin() + actionIndex );
    removeAction( actionIndex );
    --ACTIONS;
  }
}


void ActionValueFunction::setAction( const int actionIndex, const AbstractAction* actionPtr )
{
  if ( (actionIndex < 0) || (actionIndex >= ACTIONS) ) {
    RL_LOGERROR( "actionIndex " << actionIndex << " out of bounds [0, " << ACTIONS << ")" );
  }
  else {
    delete registeredActions[ actionIndex ];
    registeredActions[ actionIndex ] = actionPtr;
  }
}


const double ActionValueFunction::setGreedyActionIndex()
{
  double maxValue = -DBL_MAX;
  double value = 0.0;
  std::vector< int > actionIndexVector;
  for ( int a = 0; a < ACTIONS; ++a ) {
    value = getActionValue( a );
    if ( value > maxValue ) {
      maxValue = value;
      actionIndexVector.clear();
      actionIndexVector.push_back( a );
    }
    else if ( value == maxValue ) {
      actionIndexVector.push_back( a );
    }
  }

  activeActionIndex = actionIndexVector[
      utility::minMax( static_cast< int >( floor(utility::getNextDouble() * actionIndexVector.size()) ),
                       0, static_cast< int >( actionIndexVector.size() - 1 ) ) ];

  return maxValue;
}


const double ActionValueFunction::setRandomActionIndex()
{
  activeActionIndex = utility::minMax( static_cast< int >( floor(utility::getNextDouble() * ACTIONS) ), 0, ACTIONS - 1 );

  return getActionValue( activeActionIndex );
}


const double ActionValueFunction::setSoftMaxActionIndex( const double TEMPERATURE )
{
  if ( actionValues == 0 ) {
    actionValues = new double[ ACTIONS ];
  }

  double MIN_VALUE = DBL_MAX;
  double MAX_VALUE = -DBL_MAX;
  for ( int a = 0; a < ACTIONS; ++a ) {
    actionValues[a] = getActionValue( a );
    if ( actionValues[a] < MIN_VALUE ) {
      MIN_VALUE = actionValues[a];
    }
    if ( actionValues[a] > MAX_VALUE ) {
      MAX_VALUE = actionValues[a];
    }
  }

  activeActionIndex = utility::getSoftMaxIndex( ACTIONS, actionValues, TEMPERATURE, MIN_VALUE, MAX_VALUE );

  return getActionValue( activeActionIndex );
}


const int ActionValueFunction::save( const std::string& fileName ) const
{
  FILE* file = fopen( fileName.c_str(), "w+" );

  if ( file == 0 ) {
    RL_LOGERROR( "errno = " << errno << " --> " << strerror( errno ) );
    return -1;
  }

  // State speichern
  if ( statePtr->save( file ) != 0 ) {
    RL_LOGERROR( "saving state failed" );
    return -1;
  }

  // Anzahl der Aktionen
  if ( fwrite( &ACTIONS, sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << " --> " << strerror( errno ) );
    return -1;
  }

  // ValueFunction speichern
  if ( save( file ) != 0 ) {
    RL_LOGERROR( "saving valuefunction failed" );
    return -1;
  }

  if ( fclose( file ) != 0 ) {
    RL_LOGERROR( "errno = " << errno << " --> " << strerror( errno ) );
    return -1;
  }

  return 0;
}


const int ActionValueFunction::load( const std::string& fileName )
{
  FILE *file = fopen( fileName.c_str(), "r" );

  if ( file == 0 ) {
    RL_LOGERROR( "errno = " << errno << " --> " << strerror( errno ) );
    return -1;
  }

  // State laden
  if ( statePtr == 0 ) {
    statePtr = new State( file );
  }
  else if ( statePtr->load( file ) != 0 ) {
    RL_LOGERROR( "loading state failed" );
    return -1;
  }

  // die bereits registrierten Aktionen werden deleted
  for ( int a = 0; a < ACTIONS; ++a ) {
    delete registeredActions[a];
  }
  registeredActions.clear();

  // neue Anzahl der Aktionen
  if ( fread( &ACTIONS, sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << " --> " << strerror( errno ) );
    return -1;
  }

  // registrierte Aktionen mit NULL initialisieren
  registeredActions = std::vector< const AbstractAction* >( ACTIONS );
  for ( int a = 0; a < ACTIONS; ++a ) {
    registeredActions[a] = 0;
  }

  // ValueFunction laden
  if ( load( file ) != 0 ) {
    RL_LOGERROR( "loading valuefunction failed" );
    return -1;
  }

  if ( fclose( file ) != 0 ) {
    RL_LOGERROR( "errno = " << errno << " --> " << strerror( errno ) );
    return -1;
  }

  return 0;
}
