#ifndef __ACTION_VALUE_FUNCTION_H__
#define __ACTION_VALUE_FUNCTION_H__

#include "ValueFunction.h"
#include <string>
#include <vector>

class AbstractAction;
class State;


class ActionValueFunction : public ValueFunction
{
  public:
    /**
    * Konstruktor
    * @param aStatePtr - Pointer auf den Zustand
    */
    ActionValueFunction( State* aStatePtr );

    /**
    * Destruktor
    */
    virtual ~ActionValueFunction();

    /**
    * Registriert eine Aktion.
    * Muss unbedingt vor registerTiling (u.ae.) aufgerufen werden.
    * @param actionPtr - Pointer auf eine AbstractAction
    */
    virtual void registerAction( const AbstractAction* actionPtr );

    /**
    * Unregistered eine Aktion.
    * @param actionPtr - Pointer auf eine AbstractAction
    */
    virtual void unregisterAction( const AbstractAction* actionPtr );

    /**
    * Unregistered eine Aktion.
    * @param actionIndex - Index einer AbstractAction
    */
    virtual void unregisterAction( const int actionIndex );

    /**
    * Setzt eine Aktion. Die bisherige Aktion wird deleted.
    * @param actionIndex - der Index der Aktion die ausgetauscht werden soll
    * @param actionPtr - die neue Aktion
    */
    virtual void setAction( const int actionIndex, const AbstractAction* actionPtr );

    /**
    * Setzt den Index der Aktion mit dem hoechsten Wert zum aktuellen Zustand.
    * Gibt es mehrere Aktionen mit gleich hohem Wert, so wird eine von ihnen gleichverteilt ausgewaehlt.
    * @return der Wert der Aktion im aktuellen Zustands
    */
    virtual const double setGreedyActionIndex();

    /**
    * Setzt einen zufaelligen Index einer Aktion.
    * @return der Wert der Aktion im aktuellen Zustands
    */
    virtual const double setRandomActionIndex();

    /**
    * Setzt fuer den aktuellen Zustand eine Aktion, mit Hilfe der Boltzmannverteilung.
    * @return der Wert der Aktion im aktuellen Zustands
    */
    virtual const double setSoftMaxActionIndex( const double TEMPERATURE );

    /**
    * Liefert den Wert einer Aktion fuer den aktuellen Zustand.
    * @param der Index der Aktion
    * @return der Wert der Aktion
    */
    virtual const double getActionValue( const int actionIndex ) const = 0;

    /**
    * Abspeichern
    * @param fileName - Name der Datei die geschrieben werden soll
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int save( const std::string& fileName ) const;

    /**
    * Laden
    * Die registrierten Aktionen werden dabei geloescht und muessen mit setAction
    * neu gesetzt werden.
    * @param fileName - Name der Datei die geladen werden soll
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int load( const std::string& fileName );

    /**
    * Liefert den aktiven Aktionsindex.
    * @return activeActionIndex
    */
    inline const int getActiveActionIndex() const;

    /**
    * Liefert die aktive Aktion.
    * @return die Aktion zum activeActionIndex.
    */
    inline const AbstractAction* getActiveAction() const;

    /**
    * Liefert die Aktion mit Index index.
    * index - der Index der Aktion die geliefert werden soll
    * @return die Aktion zum index
    */
    inline const AbstractAction* getAction( int index ) const;

    /**
    * Liefert die Anzahl der moeglichen Aktionen.
    */
    inline const int getNumberOfActions() const;


  protected:
    /**
    * Entfernt die Aktion aus der ActionValueFunction und reduziert die Gesamtanzahl der Aktionen.
    * @param der Index der Aktion die entfernt werden soll
    */
    virtual void removeAction( const int actionIndex ) = 0;

    /**
    * Abspeichern
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int save( FILE* file ) const = 0;

    /**
    * Laden
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int load( FILE* file ) = 0;


    /** Anzahl der Aktionen (kann leider nicht const sein wegen laden) */
    int ACTIONS;

    /** Der aktive Aktionsindex, d.h. der Index der Aktion mit dem hoechsten Wert. */
    int activeActionIndex;

    /** die registrierten Actions */
    std::vector< const AbstractAction* > registeredActions;

    /** Werte der einzelnen Aktionen im aktuellen Zustand
    *   --> fuer SoftMax, damit nicht in jedem step Speicher alloziiert und freigegeben wird
    */
    double* actionValues;
};


inline const int ActionValueFunction::getActiveActionIndex() const
{
  return activeActionIndex;
}

inline const AbstractAction* ActionValueFunction::getActiveAction() const
{
  if ( (activeActionIndex < 0) || (activeActionIndex >= ACTIONS) ) {
    return 0;
  }
  return registeredActions[ activeActionIndex ];
}

inline const AbstractAction* ActionValueFunction::getAction( int index ) const
{
  if ( (index < 0) || (index >= ACTIONS) ) {
    return 0;
  }
  return registeredActions[ index ];
}

inline const int ActionValueFunction::getNumberOfActions() const
{
  return ACTIONS;
}

#endif
