#include "LookupTable.h"
#include "State.h"
#include "Utility.h"

#include <cerrno>
#include <cfloat>


LookupTable::LookupTable( State* aStatePtr )
  : statePtr( aStatePtr )
{
  // am Anfang haben wir keine Gewichte
  ARRAY_SIZE = 0;
  // Gewichtsvektor ist erstmal null --> wird durch create erzeugt
  weights = 0;
  // Traces erstmal null
  eTraces = 0;

  // indexOffset wird in StateLookupTable bzw. ActionLookupTable mit Werten gefuellt
  indexOffset = 0;

  // kein Zustand gesetzt
  activeIndex = -1;
  storedActiveIndex = -1;

  // Art der Speicherung der Traces
  traceStorage = Types::MAP;
}


LookupTable::~LookupTable()
{
  freeMemory();
}


void LookupTable::initWeights( const double (*initFunctionPtr)() )
{
  if ( weights != 0 ) {
    for ( int w = 0; w < ARRAY_SIZE; ++w ) {
      weights[w] = (*initFunctionPtr)();
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }
}


void LookupTable::initEligibilityTraces()
{
  if ( traceStorage == Types::ARRAY ) {
    if ( eTraces == 0 ) {
      // EligibilityTraces sind noch nicht angelegt --> Speicher holen
      eTraces = allocateMemoryWeights();
    }

    for ( int e = 0; e < ARRAY_SIZE; ++e ) {
      eTraces[e] = 0.0;
    }
  }
  else if ( traceStorage == Types::MAP ) {
    clearTraceMap();
  }
  else {
    RL_LOGERROR( "Unknown traceStorage = " << traceStorage );
  }
}


void LookupTable::updateWeights( const double ALPHA, const double DELTA )
{
  if ( weights != 0 ) {
    const double FACTOR = ALPHA * DELTA;

    if ( traceStorage == Types::ARRAY ) {
      if ( eTraces != 0 ) {
        for ( int w = 0; w < ARRAY_SIZE; ++w ) {
          weights[w] += FACTOR * eTraces[w];
        }
      }
      else if ( storedActiveIndex != -1 ) {
        weights[ storedActiveIndex ] += FACTOR;
      }
      else {
        RL_LOGERROR( "traceStorage == ARRAY and eTraces == 0 and storedActiveIndex == -1"
                     << " --> Set traceStorage to MAP"
                     << " or call initEligibilityTraces or storeState before calling updateWeights!" );
      }
    }
    else if ( traceStorage == Types::MAP ) {
      if ( eTracesMap.empty() == false ) {
        for ( std::map< int, double >::const_iterator i = eTracesMap.begin();
            i != eTracesMap.end(); ++i )
        {
          weights[ i->first ] += FACTOR * i->second;
        }
      }
      else if ( storedActiveIndex != -1 ) {
        weights[ storedActiveIndex ] += FACTOR;
      }
      else {
        RL_LOGERROR( "traceStorage == MAP and eTracesMap == empty and storedActiveIndex == -1"
                     << " --> Set traceStorage to ARRAY"
                     << " or call initEligibilityTraces or storeState before calling updateWeights!" );
      }
    }
    else {
      RL_LOGERROR( "Unknown traceStorage = " << traceStorage );
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }
}


void LookupTable::updateEligibilityTraces( const double GAMMA, const double LAMBDA )
{
  const double FACTOR = GAMMA * LAMBDA;

  if ( traceStorage == Types::ARRAY ) {
    if ( eTraces == 0 ) {
      RL_LOGERROR( "traceStorage == ARRAY and eTraces == 0"
                   << " --> Set traceStorage to MAP or call initEligibilityTraces before calling updateEligibilityTraces!" );
      return;
    }

    for ( int e = 0; e < ARRAY_SIZE; ++e ) {
      eTraces[e] *= FACTOR;
    }
  }
  else if ( traceStorage == Types::MAP ) {
    for ( std::map< int, double >::iterator i = eTracesMap.begin();
          i != eTracesMap.end(); ++i )
    {
      i->second *= FACTOR;
    }
  }
  else {
    RL_LOGERROR( "Unknown traceStorage = " << traceStorage );
  }
}


void LookupTable::checkWeightLimit( const double MIN_WEIGHT, const double MAX_WEIGHT )
{
  if ( weights != 0 ) {
    double* weightPtr = &weights[0];
    for ( int w = 0; w < ARRAY_SIZE; ++w ) {
      *weightPtr = utility::minMax( *weightPtr, MIN_WEIGHT, MAX_WEIGHT );
      ++weightPtr;
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }
}


const double LookupTable::getMinWeight() const
{
  double min = DBL_MAX;

  if ( weights != 0 ) {
    double* weightPtr = &weights[0];
    for ( int w = 0; w < ARRAY_SIZE; ++w ) {
      if ( *weightPtr < min ) {
        min = *weightPtr;
      }
      ++weightPtr;
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }

  return min;
}


const double LookupTable::getAvgWeight( const bool abs ) const
{
  if ( weights != 0 ) {
    double sum = 0.0;
    double* weightPtr = &weights[0];
    if ( abs ) {
      for ( int w = 0; w < ARRAY_SIZE; ++w ) {
        sum += fabs( *weightPtr );
        ++weightPtr;
      }
    }
    else {
      for ( int w = 0; w < ARRAY_SIZE; ++w ) {
        sum += *weightPtr;
        ++weightPtr;
      }
    }

    return (sum / ARRAY_SIZE);
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }

  return 0.0;
}


const double LookupTable::getMaxWeight() const
{
  double max = -1.0 * DBL_MAX;

  if ( weights != 0 ) {
    double* weightPtr = &weights[0];
    for ( int w = 0; w < ARRAY_SIZE; ++w ) {
      if ( *weightPtr > max ) {
        max = *weightPtr;
      }
      ++weightPtr;
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }

  return max;
}


void LookupTable::setState()
{
  calculateActiveIndex();
}


void LookupTable::calculateActiveIndex()
{
  if ( statePtr != 0 ) {
    activeIndex = 0;
    for ( unsigned int v = 0; v < statePtr->getNumberOfVariables(); ++v ) {
        activeIndex += indexOffset[v] * (statePtr->getValueInt( v ) - statePtr->getMinValueInt( v ));
    }

    if ( (activeIndex >= ARRAY_SIZE) || (activeIndex < 0) ) {
      RL_LOGERROR( "index (" << activeIndex << ") out ouf bounds --> [0, " << ARRAY_SIZE << ")" );
      activeIndex = -1;
    }
  }
  else {
    RL_LOGERROR( "statePtr == 0" );
    activeIndex = -1;
  }
}


void LookupTable::initEpisodeStateHistory()
{
  episodeIndexHistorySet.clear();
}


void LookupTable::valueEpisode( const double value )
{
  if ( episodeIndexHistorySet.empty() == false ) {
    if ( weights != 0 ) {
      for ( std::set< int >::const_iterator i = episodeIndexHistorySet.begin();
            i != episodeIndexHistorySet.end(); ++i )
      {
        weights[ *i ] += value;
      }
    }
    else {
      RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
    }
  }
  else {
    RL_LOGERROR( "episodeIndexHistorySet is empty! "
                 << "Did you forget to set ValueFunction::episodeStateHistory to true?" );
  }
}


const int LookupTable::save( FILE* file ) const
{
  // Art der EligibilityTraces (ARRAY, MAP)
  if ( fwrite( &traceStorage, sizeof( Types::TRACE_STORAGE ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Indexoffset der Variablen fuer weights und eTraces
  if ( fwrite( &indexOffset[0], sizeof( int ), statePtr->getNumberOfVariables(), file ) != static_cast< size_t >( statePtr->getNumberOfVariables() ) ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Die Anzahl der Gewichte
  if ( fwrite( &ARRAY_SIZE, sizeof( int ), 1, file ) != 1 ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Die zu lernenden Werte
  if ( weights != 0 ) {
    if ( fwrite( &weights[0], sizeof( double ), ARRAY_SIZE, file ) != static_cast< size_t >( ARRAY_SIZE ) ) {
        RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
      return -1;
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
    return -1;
  }

  return 0;
}


const int LookupTable::load( FILE* file )
{
  // falls schon Daten vorhanden sind --> Speicher freigeben und neu allozieren
  freeMemory();

  // Art der EligibilityTraces (ARRAY, MAP)
  if ( fread( &traceStorage, sizeof( Types::TRACE_STORAGE ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Speicher holen
  allocateMemory();

  // Indexoffset der Variablen fuer weights und eTraces
  if ( fread( &indexOffset[0], sizeof( int ), statePtr->getNumberOfVariables(), file ) != static_cast< size_t >( statePtr->getNumberOfVariables() ) ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Die Anzahl der Gewichte
  if ( fread( &ARRAY_SIZE, sizeof( int ), 1, file ) != 1 ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Speicher fuer weights holen
  weights = allocateMemoryWeights();

  // Die zu lernenden Werte
  if ( fread( &weights[0], sizeof( double ), ARRAY_SIZE, file ) != static_cast< size_t >( ARRAY_SIZE ) ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  return 0;
}


void LookupTable::allocateMemory()
{
  // indexOffset
  indexOffset = new int[ statePtr->getNumberOfVariables() ];
}


double* LookupTable::allocateMemoryWeights()
{
  // Weightvektor anlegen
  return (new double[ ARRAY_SIZE ]);
}


void LookupTable::freeMemory()
{
  if ( indexOffset != 0 ) {
    delete[] indexOffset;
    indexOffset = 0;
  }

  if ( weights != 0 ) {
    delete[] weights;
    weights = 0;
  }
  if ( eTraces != 0 ) {
    delete[] eTraces;
    eTraces = 0;
  }
}
