#ifndef __LOOKUP_TABLE_H__
#define __LOOKUP_TABLE_H__

#include "Types.h"

#include <map>
#include <set>

class State;

class LookupTable
{
  public:
    /**
    * Konstruktor
    * @param aStatePtr - ein Pointer auf den Zustand
    */
    LookupTable( State* aStatePtr );

    /**
    * Destruktor
    */
    virtual ~LookupTable();

    /**
    * Initialisiert die Gewichte
    * @param initFunctionPtr - Pointer auf die Funktion die zum Initialisieren bentzt wird
    */
    virtual void initWeights( const double (*initFunctionPtr)() );

    /**
    * Initialisieren der EligibilityTraces mit 0.0
    */
    virtual void initEligibilityTraces();

    /**
    * Updaten der Gewichte
    * @param ALPHA - die Lernrate
    * @param DELTA - die Aenderung des Gewichts
    */
    virtual void updateWeights( const double ALPHA, const double DELTA );

    /**
    * Updaten der EligibilityTraces
    * @param GAMMA - Discountrate
    * @param LAMBDA - TraceDecay
    */
    virtual void updateEligibilityTraces( const double GAMMA, const double LAMBDA );

    /**
    * Ueberprueft ob die Gewichte in den Grenzen sind.
    * minWeight <= weight <= maxWeight
    * @param MIN_WEIGHT - minimales Gewicht
    * @param MAX_WEIGHT - maximales Gewicht
    */
    virtual void checkWeightLimit( const double MIN_WEIGHT, const double MAX_WEIGHT );

    /**
    * Liefert das minimale Gewicht zurueck.
    * @return das minimale Gewicht
    */
    virtual const double getMinWeight() const;

    /**
    * Liefert den Durchschnitt aller Gewichte.
    * @param abs - wenn true wird der absolute Betrag der Gewichte aufsummiert
    * @return das durchschnittliche Gewicht
    */
    virtual const double getAvgWeight( const bool abs ) const;

    /**
    * Liefert das maximale Gewicht zurueck.
    * @return das maximale Gewicht
    */
    virtual const double getMaxWeight() const;


    /**
    * Setzt den Zustand --> berechnet den aktiven Index
    */
    virtual void setState();

    /**
    * Berechnet den aktiven Index.
    */
    virtual void calculateActiveIndex();

    /**
    * Initialisiert die episodeStateHistory.
    */
    virtual void initEpisodeStateHistory();

    /**
    * Bewertet die gesamte Episode.
    * --> dafuer muss episodeStateHistory auf true gesetzt sein
    * @param value - dieser Wert wird auf alle Gewichte der Episode gerechnet
    */
    virtual void valueEpisode( const double value );


    /**
    * Setzt die Art wie die Traces gehalten werden sollen.
    */
    inline void setTraceStorage( const Types::TRACE_STORAGE newTraceStorage );


  protected:
    /**
    * Abspeichern
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int save( FILE* file ) const;

    /**
    * Laden
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int load( FILE* file );

    /** Speicher fuer die Arrays holen. */
    virtual void allocateMemory();

    /**
    * Speicher fuer weights holen.
    * Da sich die Groesse des Arrays aus den anderen Werten errechnet, gibt
    * es 2 allocate Methoden.
    */
    virtual double* allocateMemoryWeights();

    /** Speicher freigeben */
    virtual void freeMemory();

    /**
    * Leert die TraceMap.
    */
    inline void clearTraceMap();

    /**
    * Setzt den StatePtr.
    */
    inline void setStatePtr( State* aStatePtr );


    /** Wie werden die Traces gespeichert Array oder Map? */
    Types::TRACE_STORAGE traceStorage;

    /** Die Anzahl der Gewichte --> darf nicht konstant sein wegen laden */
    int ARRAY_SIZE;
    /** Die Gewichte.*/
    double* weights;
    /** Die Eligibilitytraces --> Array */
    double* eTraces;
    /** Die Eligibilitytraces --> Map */
    std::map< int, double > eTracesMap;

    /**
    * Indexoffset der Variablen fuer weights und eTraces
    * --> damit der Index berechnet werden kann
    */
    int* indexOffset;

    /** Der aktive Index / Zustand */
    int activeIndex;
    /** Speichert den aktivierten Index / Zustand */
    int storedActiveIndex;

    /** Speichert alle Indizees dieser Episode */
    std::set< int > episodeIndexHistorySet;

    /** Pointer auf den Zustand */
    State* statePtr;
};


inline void LookupTable::setTraceStorage( const Types::TRACE_STORAGE newTraceStorage )
{
  traceStorage = newTraceStorage;
}

inline void LookupTable::clearTraceMap()
{
  eTracesMap.clear();
}

inline void LookupTable::setStatePtr( State* aStatePtr )
{
  statePtr = aStatePtr;
}

#endif
