#include "NeuralNetwork.h"
#include "State.h"
#include "Utility.h"

#include <cerrno>
#include <cfloat>
#include <cmath>


NeuralNetwork::NeuralNetwork( State* aStatePtr, const int numberOfHiddenNeurons )
{
  init();
  
  // der Zustand
  statePtr = aStatePtr;
  
  // Anzahl der Neuronen im InputLayer --> wird aus dem StatePtr ermittelt + Bias
  INPUT_LAYER_NEURONS = statePtr->getNumberOfVariables() + 1;
  // Anzahl der Neuronen im HiddenLayer --> wird vom User gesetzt + Bias
  HIDDEN_LAYER_NEURONS = numberOfHiddenNeurons + 1;
  // Anzahl der Neuronen im OutputLayer --> sollte 1 sein
  OUTPUT_LAYER_NEURONS = 1;
}


NeuralNetwork::NeuralNetwork()
{
  init();
}


NeuralNetwork::~NeuralNetwork()
{
  // Speicher freigeben
  freeMemory();
}


void NeuralNetwork::init()
{
  // Anzahl der Neuronen im InputLayer --> wird aus dem StatePtr ermittelt + Bias
  INPUT_LAYER_NEURONS = 0;
  // Anzahl der Neuronen im HiddenLayer --> wird vom User gesetzt + Bias
  HIDDEN_LAYER_NEURONS = 0;
  // Anzahl der Neuronen im OutputLayer --> sollte 1 sein
  OUTPUT_LAYER_NEURONS = 0;

  // die Werte im Inputlayer
  inputValues = 0;
  // die Werte im Hiddenlayer
  hiddenValues = 0;
  // die Werte im Outputlayer
  outputValues = 0;

  // die Gewichte vom Inputlayer zum Hiddenlayer
  inputHiddenWeights = 0;
  // die Gewichte vom Hiddenlayer zum Outputlayer
  hiddenOutputWeights = 0;
  
  // Traces sind 0 --> beim ersten benutzen initialisieren
  inputHiddenOutputTraces = 0;
  hiddenOutputTraces = 0;

  // tmp-array
  tmp = 0;

  // Skalierungsfaktoren
  scaleFactor = 0;
}


void NeuralNetwork::initWeights( const double (*initFunctionPtr)() )
{
  if ( (inputHiddenWeights != 0) && (hiddenOutputWeights != 0) ) {
    // input-hidden
    double* weightPtr = &inputHiddenWeights[0];
    for ( int i = 0; i < INPUT_HIDDEN_ARRAY_SIZE; ++i ) {
      *weightPtr = (*initFunctionPtr)();
      ++weightPtr;
    }
    // hidden-output
    weightPtr = &hiddenOutputWeights[0];
    for ( int i = 0; i < HIDDEN_OUTPUT_ARRAY_SIZE; ++i ) {
      *weightPtr = (*initFunctionPtr)();
      ++weightPtr;
    }
  }
  else {
    if ( inputHiddenWeights == 0 ) {
      RL_LOGERROR( "inputHiddenWeights == 0 --> This should never happen!" );
    }
    if ( hiddenOutputWeights == 0 ) {
      RL_LOGERROR( "hiddenOutputWeights == 0 --> This should never happen!" );
    }
  }
}


void NeuralNetwork::initEligibilityTraces()
{
  if ( inputHiddenOutputTraces == 0 ) {
    inputHiddenOutputTraces = new double[ INPUT_HIDDEN_OUTPUT_ARRAY_SIZE ];
  }
  double* tracePtr = &inputHiddenOutputTraces[0];
  for ( int i = 0; i < INPUT_HIDDEN_OUTPUT_ARRAY_SIZE; ++i ) {
    *tracePtr = 0.0;
    ++tracePtr;
  }

  if ( hiddenOutputTraces == 0 ) {
    hiddenOutputTraces = new double[ HIDDEN_OUTPUT_ARRAY_SIZE ];
  }
  tracePtr = &hiddenOutputTraces[0];
  for ( int i = 0; i < HIDDEN_OUTPUT_ARRAY_SIZE; ++i ) {
    *tracePtr = 0.0;
    ++tracePtr;
  }
}


void NeuralNetwork::updateWeights( const double ALPHA, const double DELTA, const double minWeightLimit, const double maxWeightLimit )
{
  // berechnet delta_t_i fuer i aus Output
  double* outputValuesPtr = &outputValues[0];
  double* tmpPtr = &tmp[0];
  for ( int o = 0; o < OUTPUT_LAYER_NEURONS; ++o ) {
    *tmpPtr = DELTA * *outputValuesPtr * (1.0 - *outputValuesPtr);
    ++tmpPtr;
    ++outputValuesPtr;
  }

  // hidden - output
  if ( hiddenOutputWeights != 0 ) {
    double* weightPtr = &hiddenOutputWeights[0];
    if ( hiddenOutputTraces == 0 ) {
      // ohne Traces
      double* hiddenValuesPtr = &hiddenValues[0];
      for ( int h = 0; h < HIDDEN_LAYER_NEURONS; ++h ) {
        tmpPtr = &tmp[0];
        for ( int o = 0; o < OUTPUT_LAYER_NEURONS; ++o ) {
          *weightPtr += ALPHA * *tmpPtr * *hiddenValuesPtr;
          ++weightPtr;
        }
        ++hiddenValuesPtr;
      }
    }
    else {
      // mit Traces
      double* tracePtr = &hiddenOutputTraces[0];
      for ( int i = 0; i < HIDDEN_LAYER_NEURONS * OUTPUT_LAYER_NEURONS; ++i ) {
        *weightPtr += ALPHA * DELTA * *tracePtr;
        ++weightPtr;
        ++tracePtr;
      }
    }
  }
  else {
    RL_LOGERROR( "hiddenOutputWeights == 0 --> This should never happen!" );
  }

  // input - hidden
  if ( inputHiddenWeights != 0 ) {
    double* weightPtr = &inputHiddenWeights[0];
    if ( inputHiddenOutputTraces == 0 ) {
      // ohne Traces
      double* inputValuesPtr = &inputValues[0];
      for ( int i = 0; i < INPUT_LAYER_NEURONS; ++i ) {
        double* hiddenValuesPtr = &hiddenValues[0];
        double* hiddenOutputWeightsPtr = &hiddenOutputWeights[0];
        for ( int h = 0; h < (HIDDEN_LAYER_NEURONS - 1); ++h ) {
          double d = 0.0;
          tmpPtr = &tmp[0];
          for ( int o = 0; o < OUTPUT_LAYER_NEURONS; ++o ) {
            d += *tmpPtr * *hiddenOutputWeightsPtr * *hiddenValuesPtr * (1.0 - *hiddenValuesPtr);
            ++tmpPtr;
            ++hiddenOutputWeightsPtr;
          }
          *weightPtr += ALPHA * d * *inputValuesPtr;
          ++weightPtr;
          ++hiddenValuesPtr;
        }
        ++inputValuesPtr;
      }
    }
    else {
      // mit Traces
      double* tracePtr = &inputHiddenOutputTraces[0];
      for ( int i = 0; i < INPUT_LAYER_NEURONS * (HIDDEN_LAYER_NEURONS - 1); ++i ) {
        double sum = 0.0;
        for ( int o = 0; o < OUTPUT_LAYER_NEURONS; ++o ) {
          sum += DELTA * *tracePtr;
          ++tracePtr;
        }
        *weightPtr += ALPHA * sum;
        ++weightPtr;
      }
    }
  }
  else {
    RL_LOGERROR( "inputHiddenWeights == 0 --> This should never happen!" );
  }

  // die Grenzen der Gewichte ueberpruefen
  if ( (minWeightLimit > -1.0 * DBL_MAX) || (maxWeightLimit < DBL_MAX) ) {
    checkWeightLimit(minWeightLimit, maxWeightLimit);
  }
}


void NeuralNetwork::updateEligibilityTraces( const double GAMMA, const double LAMBDA )
{
  if ( outputValues != 0 ) {
    double* outputValuesPtr = &outputValues[0];
    double* tmpPtr = &tmp[0];
    for ( int o = 0; o < OUTPUT_LAYER_NEURONS; ++o ) {
      *tmpPtr = *outputValuesPtr * (1.0 - *outputValuesPtr);
      ++outputValuesPtr;
      ++tmpPtr;
    }
  }
  else {
    RL_LOGERROR( "outputValues == 0 --> This should never happen!" );
  }

  // hidden - output
  if ( hiddenOutputTraces != 0 ) {
    double* tracePtr = &hiddenOutputTraces[0];
    double* hiddenValuesPtr = &hiddenValues[0];
    for ( int h = 0; h < HIDDEN_LAYER_NEURONS; ++h ) {
      double* tmpPtr = &tmp[0];
      for ( int o = 0; o < OUTPUT_LAYER_NEURONS; ++o ) {
        *tracePtr = LAMBDA * *tracePtr + *tmpPtr * *hiddenValuesPtr;
        ++tracePtr;
        ++tmpPtr;
      }
      ++hiddenValuesPtr;
    }
  }
  else {
    RL_LOGERROR( "hiddenOutputTraces == 0" );
  }

  // input - hidden - output
  if ( inputHiddenOutputTraces != 0 ) {
    double* inputHiddenOutputTracesPtr = &inputHiddenOutputTraces[0];
    double* inputValuesPtr = &inputValues[0];
    for ( int i = 0; i < INPUT_LAYER_NEURONS; ++i ) {
      double* hiddenOutputWeightsPtr = &hiddenOutputWeights[0];
      double* hiddenValuesPtr = &hiddenValues[0];
      for ( int h = 0; h < (HIDDEN_LAYER_NEURONS - 1); ++h ) { // -1 wegen Bias
        double* tmpPtr = &tmp[0];
        for ( int o = 0; o < OUTPUT_LAYER_NEURONS; ++o ) {
          *inputHiddenOutputTracesPtr = LAMBDA * *inputHiddenOutputTracesPtr
                                        + *tmpPtr * *hiddenOutputWeightsPtr * *hiddenValuesPtr * (1.0 - *hiddenValuesPtr) * *inputValuesPtr;
          ++inputHiddenOutputTracesPtr;
          ++hiddenOutputWeightsPtr;
          ++tmpPtr;
        }
        ++hiddenValuesPtr;
      }
      ++inputValuesPtr;
    }
  }
  else {
    RL_LOGERROR( "inputHiddenOutputTraces == 0" );
  }
}


void NeuralNetwork::accumulatingTraces( const double VALUE )
{
  // nichts tun --> passiert alles in updateEligibilityTraces
}


void NeuralNetwork::replacingTraces( const double VALUE )
{
  // nichts tun --> passiert alles in updateEligibilityTraces
}


const double NeuralNetwork::getOutputValue(const int outputIndex) const
{
  if ((outputIndex < 0) || (outputIndex >= OUTPUT_LAYER_NEURONS)) {
    RL_LOGERROR("outputIndex out of Bounds --> 0 <= " << outputIndex << " < " << OUTPUT_LAYER_NEURONS);
    return 0.0;
  }

  // hidden
  // init --> hidden-value auf 0
  double* hiddenValuesPtr = &hiddenValues[0];
  for ( int h = 0; h < (HIDDEN_LAYER_NEURONS - 1); ++h ) {
    *hiddenValuesPtr = 0.0;
    ++hiddenValuesPtr;
  }
  // hidden-value berechnen
  double* inputValuesPtr = &inputValues[0];
  double* inputHiddenWeightsPtr = &inputHiddenWeights[0];
  for ( int i = 0; i < INPUT_LAYER_NEURONS; ++i ) {
    hiddenValuesPtr = &hiddenValues[0];
    for ( int h = 0; h < (HIDDEN_LAYER_NEURONS - 1); ++h ) { // -1 wegen Bias
      *hiddenValuesPtr += *inputValuesPtr * *inputHiddenWeightsPtr;
      ++hiddenValuesPtr;
      ++inputHiddenWeightsPtr;
    }
    ++inputValuesPtr;
  }
  // sigmoide
  hiddenValuesPtr = &hiddenValues[0];
  for ( int h = 0; h < (HIDDEN_LAYER_NEURONS - 1); ++h ) {
    *hiddenValuesPtr = 1.0 / (1.0 + exp( -1.0 * *hiddenValuesPtr )); /* asymmetric sigmoid */
    ++hiddenValuesPtr;
  }

  // output
  // init --> output-value auf 0
  double* outputValuesPtr = &outputValues[0];
  for ( int o = 0; o < OUTPUT_LAYER_NEURONS; ++o ) {
    *outputValuesPtr = 0.0;
    ++outputValuesPtr;
  }
  // output-value berechnen
  hiddenValuesPtr = &hiddenValues[0];
  double* hiddenOutputWeightsPtr = &hiddenOutputWeights[0];
  for ( int h = 0; h < HIDDEN_LAYER_NEURONS; ++h ) {
    outputValuesPtr = &outputValues[0];
    for ( int o = 0; o < OUTPUT_LAYER_NEURONS; ++o ) {
      *outputValuesPtr += *hiddenValuesPtr * *hiddenOutputWeightsPtr;
      ++outputValuesPtr;
      ++hiddenOutputWeightsPtr;
    }
    ++hiddenValuesPtr;
  }
  // sigmoide
  outputValuesPtr = &outputValues[0];
  for ( int o = 0; o < OUTPUT_LAYER_NEURONS; ++o ) {
    *outputValuesPtr = 1.0 / (1.0 + exp( -1.0 * *outputValuesPtr )); /* asymmetric sigmoid (OPTIONAL) */
    ++outputValuesPtr;
  }

  return outputValues[outputIndex];
}


void NeuralNetwork::checkWeightLimit( const double minWeightLimit, const double maxWeightLimit )
{
  // hidden - output
  if ( hiddenOutputWeights != 0 ) {
    double* weightPtr = &hiddenOutputWeights[0]; // TODO: nur fuer die aktuelle Action ueberpruefen
    for ( int i = 0; i < HIDDEN_OUTPUT_ARRAY_SIZE; ++i ) {
      *weightPtr = utility::minMax( *weightPtr, minWeightLimit, maxWeightLimit );
      ++weightPtr;
    }
  }
  else {
    RL_LOGERROR( "hiddenOutputWeights == 0 --> This should never happen!" );
  }

  // input - hidden
  if ( inputHiddenWeights != 0 ) {
    double* weightPtr = &inputHiddenWeights[0]; // TODO: nur fuer die aktuelle Action ueberpruefen
    for ( int i = 0; i < INPUT_HIDDEN_ARRAY_SIZE; ++i ) {
      *weightPtr = utility::minMax( *weightPtr, minWeightLimit, maxWeightLimit );
      ++weightPtr;
    }
  }
  else {
    RL_LOGERROR( "inputHiddenWeights == 0 --> This should never happen!" );
  }
}


const double NeuralNetwork::getMinWeight() const
{
  double min = DBL_MAX;

  // hidden - output
  if ( hiddenOutputWeights != 0 ) {
    double* weightPtr = &hiddenOutputWeights[0];
    for ( int i = 0; i < HIDDEN_OUTPUT_ARRAY_SIZE; ++i ) {
      if ( *weightPtr < min ) {
        min = *weightPtr;
      }
      ++weightPtr;
    }
  }
  else {
    RL_LOGERROR( "hiddenOutputWeights == 0 --> This should never happen!" );
  }

  // input - hidden
  if ( inputHiddenWeights != 0 ) {
    double* weightPtr = &inputHiddenWeights[0];
    for ( int i = 0; i < INPUT_HIDDEN_ARRAY_SIZE; ++i ) {
      if ( *weightPtr < min ) {
        min = *weightPtr;
      }
      ++weightPtr;
    }
  }
  else {
    RL_LOGERROR( "inputHiddenWeights == 0 --> This should never happen!" );
  }

  return min;
}


const double NeuralNetwork::getAvgWeight( const bool abs ) const
{
  double sum = 0.0;

  // hidden - output
  if ( hiddenOutputWeights != 0 ) {
    double* weightPtr = &hiddenOutputWeights[0];
    if ( abs ) {
      for ( int i = 0; i < HIDDEN_OUTPUT_ARRAY_SIZE; ++i ) {
        sum += fabs( *weightPtr );
        ++weightPtr;
      }
    }
    else {
      for ( int i = 0; i < HIDDEN_OUTPUT_ARRAY_SIZE; ++i ) {
        sum += *weightPtr;
        ++weightPtr;
      }
    }
  }
  else {
    RL_LOGERROR( "hiddenOutputWeights == 0 --> This should never happen!" );
  }

  // input - hidden
  if ( inputHiddenWeights != 0 ) {
    double* weightPtr = &inputHiddenWeights[0];
    if ( abs ) {
      for ( int i = 0; i < INPUT_HIDDEN_ARRAY_SIZE; ++i ) {
        sum += fabs( *weightPtr );
        ++weightPtr;
      }
    }
    else {
      for ( int i = 0; i < INPUT_HIDDEN_ARRAY_SIZE; ++i ) {
        sum += *weightPtr;
        ++weightPtr;
      }
    }
  }
  else {
    RL_LOGERROR( "inputHiddenWeights == 0 --> This should never happen!" );
  }

  return ( sum / (HIDDEN_OUTPUT_ARRAY_SIZE * INPUT_HIDDEN_ARRAY_SIZE) );
}


const double NeuralNetwork::getMaxWeight() const
{
  double max = -1.0 * DBL_MAX;

  // hidden - output
  if ( hiddenOutputWeights != 0 ) {
    double* weightPtr = &hiddenOutputWeights[0];
    for ( int i = 0; i < HIDDEN_OUTPUT_ARRAY_SIZE; ++i ) {
      if ( *weightPtr > max ) {
        max = *weightPtr;
      }
      ++weightPtr;
    }
  }
  else {
    RL_LOGERROR( "hiddenOutputWeights == 0 --> This should never happen!" );
  }

  // input - hidden
  if ( inputHiddenWeights != 0 ) {
    double* weightPtr = &inputHiddenWeights[0];
    for ( int i = 0; i < INPUT_HIDDEN_ARRAY_SIZE; ++i ) {
      if ( *weightPtr > max ) {
        max = *weightPtr;
      }
      ++weightPtr;
    }
  }
  else {
    RL_LOGERROR( "inputHiddenWeights == 0 --> This should never happen!" );
  }

  return max;
}


void NeuralNetwork::setState()
{
  for ( int i = 0; i < INPUT_LAYER_NEURONS - 1; ++i ) { // - Bias
    inputValues[i] = scaleFactor[i] * statePtr->getValueDbl( i );
  }
}


void NeuralNetwork::storeState()
{
  // nix
}


void NeuralNetwork::initEpisodeStateHistory()
{
  // Nicht implementiert, da ich nicht sicher bin ob das Sinn macht.
  // Kein RL_LOGERROR, da sonst zum Episodenstart immer Fehlerausschrift.
}


void NeuralNetwork::valueEpisode( const double value )
{
  // Ich bin nicht sicher ob das ueberhaupt Sinn macht fuer NN.
  RL_LOGERROR( "Not implemented for NeuralNetwork!" );
}


const int NeuralNetwork::save( FILE* file ) const
{
  // Anzahl der Neuronen im InputLayer
  if ( fwrite( &INPUT_LAYER_NEURONS, sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Anzahl der Neuronen im HiddenLayer
  if ( fwrite( &HIDDEN_LAYER_NEURONS, sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Anzahl der Neuronen im OutputLayer
  if ( fwrite( &OUTPUT_LAYER_NEURONS, sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // die Gewichte vom Inputlayer zum Hiddenlayer
  if ( fwrite( &inputHiddenWeights[0], sizeof( double ), INPUT_HIDDEN_ARRAY_SIZE, file ) != static_cast< unsigned int >( INPUT_HIDDEN_ARRAY_SIZE ) ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // die Gewichte vom Hiddenlayer zum Outputlayer
  if ( fwrite( &hiddenOutputWeights[0], sizeof( double ), HIDDEN_OUTPUT_ARRAY_SIZE, file ) != static_cast< unsigned int >( HIDDEN_OUTPUT_ARRAY_SIZE ) ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // die Skalierungsfaktoren
  if ( fwrite( &scaleFactor[0], sizeof( double ), INPUT_LAYER_NEURONS - 1, file ) != static_cast< unsigned int >( INPUT_LAYER_NEURONS - 1 ) ) { // - Bias
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }
  
  // Bias - Input
  if ( fwrite( &inputValues[ INPUT_LAYER_NEURONS - 1 ], sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }
  // Bias - Hidden
  if ( fwrite( &hiddenValues[ HIDDEN_LAYER_NEURONS - 1 ], sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  return 0;
}


const int NeuralNetwork::load( FILE* file )
{
  // Anzahl der Neuronen im InputLayer
  if ( fread( &INPUT_LAYER_NEURONS, sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Anzahl der Neuronen im HiddenLayer
  if ( fread( &HIDDEN_LAYER_NEURONS, sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Anzahl der Neuronen im OutputLayer
  if ( fread( &OUTPUT_LAYER_NEURONS, sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Arraygroessen berechnen und Speicher alloziieren
  create();

  // die Gewichte vom Inputlayer zum Hiddenlayer
  if ( fread( &inputHiddenWeights[0], sizeof( double ), INPUT_HIDDEN_ARRAY_SIZE, file ) != static_cast< unsigned int >( INPUT_HIDDEN_ARRAY_SIZE ) ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // die Gewichte vom Hiddenlayer zum Outputlayer
  if ( fread( &hiddenOutputWeights[0], sizeof( double ), HIDDEN_OUTPUT_ARRAY_SIZE, file ) != static_cast< unsigned int >( HIDDEN_OUTPUT_ARRAY_SIZE ) ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // die Skalierungsfaktoren
  if ( fread( &scaleFactor[0], sizeof( double ), INPUT_LAYER_NEURONS - 1, file ) != static_cast< unsigned int >( INPUT_LAYER_NEURONS - 1 ) ) { // - Bias
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }
  
  // Bias - Input
  if ( fread( &inputValues[ INPUT_LAYER_NEURONS - 1 ], sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }
  // Bias - Hidden
  if ( fread( &hiddenValues[ HIDDEN_LAYER_NEURONS - 1 ], sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  return 0;
}


void NeuralNetwork::allocateMemory()
{
  // die Werte im Inputlayer
  inputValues = new double[ INPUT_LAYER_NEURONS ];
  // die Werte im Hiddenlayer
  hiddenValues = new double[ HIDDEN_LAYER_NEURONS ];
  // die Werte im Outputlayer
  outputValues = new double[ OUTPUT_LAYER_NEURONS ];

  // die Gewichte vom Inputlayer zum Hiddenlayer
  inputHiddenWeights = new double[ INPUT_HIDDEN_ARRAY_SIZE ];
  // die Gewichte vom Hiddenlayer zum Outputlayer
  hiddenOutputWeights = new double[ HIDDEN_OUTPUT_ARRAY_SIZE ];

  // tmp-array
  tmp = new double[ OUTPUT_LAYER_NEURONS ];

  // Skalierungsfaktoren
  scaleFactor = new double[ INPUT_LAYER_NEURONS - 1 ]; // - Bias
}


void NeuralNetwork::freeMemory()
{
  // werte
  delete[] inputValues;
  delete[] hiddenValues;
  delete[] outputValues;
  // weights
  delete[] inputHiddenWeights;
  delete[] hiddenOutputWeights;
  // traces
  delete[] inputHiddenOutputTraces;
  delete[] hiddenOutputTraces;
  // tmp-array
  delete[] tmp;
  // Skalierungsfaktoren
  delete[] scaleFactor;
}
