#ifndef __NEURAL_NETWORK_H__
#define __NEURAL_NETWORK_H__

#include <list>
#include <string>

class State;


class NeuralNetwork
{
  public:
    /**
    * Konstruktor
    * @param aStatePtr - Pointer auf den Zustand
    * @param numberOfHiddenNeurons - Anzahl der Neuronen im Hiddenlayer
    */
    NeuralNetwork( State* aStatePtr, const int numberOfHiddenNeurons );

    /**
    * Konstruktor zum Laden
    */
    NeuralNetwork();

    /** Destruktor */
    virtual ~NeuralNetwork();

    /**
    * Erzeugt den Gewichtsvektor
    */
    virtual void create() = 0;
    
    /**
    * Initialisiert den Scale-Factor und Bias.
    */
    virtual void init();

    /**
    * Initialisiert die Gewichte
    * @param initFunctionPtr - Pointer auf die Funktion die zum Initialisieren bentzt wird
    */
    virtual void initWeights( const double (*initFunctionPtr)() );

    /**
    * Initialisieren der EligibilityTraces mit 0.0
    */
    virtual void initEligibilityTraces();

    /**
    * Updaten der Gewichte
    * @param ALPHA - die Lernrate
    * @param DELTA - die Aenderung des Gewichts
    * @param minWeightLimit - kleinstes erlaubtes Gewicht
    * @param maxWeightLimit - groesstes erlaubtes Gewicht
    */
    virtual void updateWeights( const double ALPHA, const double DELTA, const double minWeightLimit, const double maxWeightLimit );

    /**
    * Updaten der EligibilityTraces
    * @param GAMMA - Discountrate
    * @param LAMBDA - TraceDecay
    */
    virtual void updateEligibilityTraces( const double GAMMA, const double LAMBDA );

    /**
    * In dieser Methode passiert nichts --> wird alles in updateEligibilityTraces erledigt
    * @param VALUE - der Wert um den erhoeht wird
    */
    virtual void accumulatingTraces( const double VALUE );

    /**
    * In dieser Methode passiert nichts --> wird alles in updateEligibilityTraces erledigt
    * @param VALUE - der Wert auf den gesetzt wird
    */
    virtual void replacingTraces( const double VALUE );

    /**
    * Berechnet den Output des Netzwerks
    * @param outputIndex - liefert den Output des OutputNeurons mit index outputIndex
    */
    virtual const double getOutputValue(const int outputIndex) const;

    /**
    * Ueberprueft ob die Gewichte in den Grenzen sind.
    * minWeight <= weight <= maxWeight
    * @param minWeightLimit - kleinstes erlaubtes Gewicht
    * @param maxWeightLimit - groesstes erlaubtes Gewicht
    */
    virtual void checkWeightLimit( const double minWeightLimit, const double maxWeightLimit );

    /**
    * Liefert das minimale Gewicht zurueck.
    * @return das minimale Gewicht
    */
    virtual const double getMinWeight() const;

    /**
    * Liefert das durchschnittliche Gewicht zurueck.
    * @param abs - wenn true wird der Durchschnitt ueber den absoluten Betrag der Gewichte gebildet
    * @return das durchschnittliche Gewicht
    */
    virtual const double getAvgWeight( const bool abs ) const;

    /**
    * Liefert das maximale Gewicht zurueck.
    * @return das maximale Gewicht
    */
    virtual const double getMaxWeight() const;

    /**
    * Setzt den Zustand
    */
    virtual void setState();

    /**
    * Falls keine EligibilityTraces verwendet werden, muss der Zustand gespeichert
    * werden, damit man weiss welche Gewichte geupdatet werden muessen.
    * Beim Neuronalen Netz passiert hier nichts
    */
    virtual void storeState();

    /**
    * Initialisiert die episodeStateHistory.
    * Nicht implementiert, da ich nicht sicher bin, ob das Sinn macht.
    */
    virtual void initEpisodeStateHistory();

    /**
    * Bewertet die gesamte Episode.
    * --> dafuer muss episodeStateHistory auf true gesetzt sein
    * Nicht implementiert, da ich nicht sicher bin, ob das Sinn macht.
    * @param value - dieser Wert wird auf alle Gewichte der Episode gerechnet
    */
    virtual void valueEpisode( const double value );

    /**
    * Setzt den Scalierungsfaktor fuer eine State-Variable.
    * @param varId - StateId der Variablen
    * @param factor - Skalierungsfaktor
    */
    inline void setScaleFactor( const int varId, const double factor );

    /**
    * Setzt den Bias.
    * @param biasValue - der Wert auf den der Bias gesetzt wird
    */
    inline void setBias( const double inputBiasValue, const double hiddenBiasValue );


  protected:
    /**
    * Abspeichern
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int save( FILE* file ) const;

    /**
    * Laden
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int load( FILE* file );

    /**
    * Speicher fuer die Arrays holen.
    */
    virtual void allocateMemory();

    /**
    * Speicher freigeben
    */
    virtual void freeMemory();

    /**
    * Setzt den StatePtr.
    */
    inline void setStatePtr( State* aStatePtr );


    // Anzahl der Neuronen im InputLayer --> wird aus dem StatePtr ermittelt
    int INPUT_LAYER_NEURONS;
    // Anzahl der Neuronen im HiddenLayer --> wird vom User gesetzt
    int HIDDEN_LAYER_NEURONS;
    // Anzahl der Neuronen im OutputLayer --> sollte 1 sein
    int OUTPUT_LAYER_NEURONS;
    
    // Groesse der Arrays --> abhaengig von State- bzw. ActionValueFunction
    int INPUT_HIDDEN_ARRAY_SIZE;
    int HIDDEN_OUTPUT_ARRAY_SIZE;
    int INPUT_HIDDEN_OUTPUT_ARRAY_SIZE;

    // die Werte im Inputlayer
    double* inputValues;
    // die Werte im Hiddenlayer
    double* hiddenValues;
    // die Werte im Outputlayer
    double* outputValues;

    // die Gewichte vom Inputlayer zum Hiddenlayer
    double* inputHiddenWeights;
    // die Gewichte vom Hiddenlayer zum Outputlayer
    double* hiddenOutputWeights;

    // EligibilityTraces input-hidden-output
    double* inputHiddenOutputTraces;
    // EligibilityTraces hidden-output
    double* hiddenOutputTraces;

    // tmp-array damit nicht in jedem step neuer Speicher alloziiert werden muss
    double* tmp;

    // die Skalierungsfaktoren
    double* scaleFactor;

    /** Pointer auf den Zustand */
    State* statePtr;
};


inline void NeuralNetwork::setScaleFactor( const int varId, const double factor )
{
  scaleFactor[varId] = factor;
}

inline void NeuralNetwork::setBias( const double inputBiasValue, const double hiddenBiasValue )
{
  if ( inputValues != 0 ) {
    inputValues[ INPUT_LAYER_NEURONS - 1 ] = inputBiasValue;
  }
  if ( hiddenValues != 0 ) {
    hiddenValues[ HIDDEN_LAYER_NEURONS - 1 ] = hiddenBiasValue;
  }
}

inline void NeuralNetwork::setStatePtr( State* aStatePtr )
{
  statePtr = aStatePtr;
}

#endif
