#include "RL++.h"
#include "Log.h"
#include "Utility.h"

#include <cerrno>
#include <csignal>
#include <fstream>


RL_PP::RL_PP( const std::string& fileName )
  : confFileName( fileName )
{
  // die Lernparameter initialisieren
  ALPHA = 0.1;
  LAMBDA = 0.5;
  GAMMA = 0.9;
  TRACES = REPLACING;
  TRACE_VALUE = 1.0;
  ACTION_SELECTION = GREEDY;
  EPSILON = 0.1;
  TEMPERATURE = 0.5;

  episodes = 0;
  steps = 0;

  continueLearning = true;
  continueEpisode = true;
  sigTermRcvd = false;
  sigIntRcvd = false;

  utility::setSeed( 1 );

  installSignalHandler();

  // Konfiguration laden
  loadConf();
}


RL_PP::~RL_PP()
{
  removeSignalHandler();
}


void RL_PP::controlLoop()
{
  if ( checkValueFunction() == false ) {
    RL_LOGERROR( "valueFunctionPtr not set --> leave controlLoop !!!" );
    return;
  }

  // Ein Schleifendurchlauf entspricht einer Episode
  while ( (continueLearning) && (sigTermRcvd == false) && (sigIntRcvd == false) ) {
    // episode vorbereiten
    initEpisode();

    // ein schleifendurchlauf entspricht zeitschritt in einer episode
    continueEpisode = true;
    while ( (continueEpisode) && (sigIntRcvd == false) ) {
      // ein weiterer Schritt in der Episode
      step();

      // Signale ueberpruefen
      checkPendingSignals();
    }

    // episode abschliessen
    finishEpisode();
  }
}


void RL_PP::finishEpisode()
{
  // Statistik
  RL_LOGDATA( "episode = " << episodes << "  steps = " << steps );
}


const bool RL_PP::randomAction() const
{
  return (utility::getNextDouble() < EPSILON);
}


void RL_PP::saveConf() const
{
  // file anlegen
  std::ofstream file( confFileName.c_str(), std::ios_base::out );

  if ( file.good() == false ) {
    RL_LOGDATA( "Was not able to create file (" << confFileName << ")" );
    return;
  }

  // output
  file << "ALPHA " << ALPHA << std::endl;
  file << "LAMBDA " << LAMBDA << std::endl;
  file << "GAMMA " << GAMMA << std::endl;
  file << "ACTION_SELECTION " << ((ACTION_SELECTION == GREEDY) ?         "GREEDY" :
                                 ((ACTION_SELECTION == EPSILON_GREEDY) ? "EPSILON_GREEDY" :
                                                                         "SOFTMAX"))
       << std::endl;
  file << "EPSILON " << EPSILON << std::endl;
  file << "TEMPERATURE " << TEMPERATURE << std::endl;
  file << "TRACES " << ((TRACES == REPLACING) ?    "REPLACING" :
                       ((TRACES == ACCUMULATING) ? "ACCUMULATING" :
                                                   "NONE"))
       << std::endl;
  file << "TRACE_VALUE " << TRACE_VALUE << std::endl;

  // schliessen
  file.close();
}


void RL_PP::loadConf()
{
  const unsigned int maxValueLength = 512;

  // file anlegen
  std::ifstream file( confFileName.c_str(), std::ios_base::in );

  if ( file.good() == false ) {
    RL_LOGDATA( "File (" << confFileName << ") didn't exist --> create" );
    saveConf();
    return;
  }

  // input
  std::string param, value;
  char ca[ maxValueLength ];

  while ( (file.good() == true) && (file.peek() != EOF) ) {
    char c = file.peek();

    if ( c == '#' ) {
      // Kommentar --> Zeile ueberspringen
      file.getline( ca, maxValueLength, '\n' );
    }
    else if ( (c == ' ') || (c == '\t') ) {
      // die Zeile faengt nicht gut an --> Meldung
      file.getline( ca, maxValueLength, '\n' );
      RL_LOGERROR( "spaces at begin of line: " << ca );
    }
    else {
      // Parameter einlesen
      file >> param;

      // den Separator (Space) ignorieren
      file.ignore( 1 );

      // den Wert des Parameter einlesen
      file.getline( ca, maxValueLength, '\n' );
      value = std::string( ca );

      // Wert setzen
      if ( param == "ALPHA" ) {
        ALPHA = atof( value.c_str() );
      }
      else if ( param == "LAMBDA" ) {
        LAMBDA = atof( value.c_str() );
      }
      else if ( param == "GAMMA" ) {
        GAMMA = atof( value.c_str() );
      }
      else if ( param == "ACTION_SELECTION" ) {
        if ( value == "GREEDY" ) {
          ACTION_SELECTION = GREEDY;
        }
        else if ( value == "EPSILON_GREEDY" ) {
          ACTION_SELECTION = EPSILON_GREEDY;
        }
        else if ( value == "SOFTMAX" ) {
          ACTION_SELECTION = SOFTMAX;
        }
        else {
          RL_LOGERROR( "Unknown ActionSelection Algorithm " << value );
        }
      }
      else if ( param == "EPSILON" ) {
        EPSILON = atof( value.c_str() );
      }
      else if ( param == "TEMPERATURE" ) {
        TEMPERATURE = atof( value.c_str() );
      }
      else if ( param == "TRACES" ) {
        if ( value == "REPLACING" ) {
          TRACES = REPLACING;
        }
        else if ( value == "ACCUMULATING" ) {
          TRACES = ACCUMULATING;
        }
        else if ( value == "NONE" ) {
          TRACES = NONE;
        }
        else {
          RL_LOGERROR( "Unknown Tracetype " << value );
        }
      }
      else if ( param == "TRACE_VALUE" ) {
        TRACE_VALUE = atof( value.c_str() );
      }
    }
  }

  // schliessen
  file.close();
}


void RL_PP::installSignalHandler()
{
  // die bisher geblockten Signale werden hier gespeichert --> erstmal leeren
  sigemptyset( &oldSigSet );

  // signale die geblockt werden sollen in Menge aufnehmen
  sigemptyset( &newSigSet );
  sigaddset( &newSigSet, SIGUSR1 );
  sigaddset( &newSigSet, SIGUSR2 );
  sigaddset( &newSigSet, SIGTERM );
  sigaddset( &newSigSet, SIGINT );

  // Signale in newSigSet blocken (bisher geblockte Signale werden
  // in oldSigSet gespeichert und um newSigSet erweitert)
  if ( sigprocmask( SIG_BLOCK, &newSigSet, &oldSigSet ) == -1 ){
    RL_LOGERROR( "sigprocmask --> errno = " << errno << " - " << strerror( errno ) );
  }

  // die geblockten Signale werden hier gespeichert --> Menge erstmal leeren
  sigemptyset( &pendingSignals );
}


void RL_PP::removeSignalHandler()
{
  // nur Signale unblocken die von uns hinzugefuegt wurden
  // (fuer den Fall das ein Signal in newSigSet und oldSigSet enthalten ist,
  //  das soll natuerlich auch weiterhin geblockt werden)
  if ( sigismember( &oldSigSet, SIGUSR1 ) == 1 ) {
    sigdelset( &newSigSet, SIGUSR1 );
  }
  if ( sigismember( &oldSigSet, SIGUSR2 ) == 1 ) {
    sigdelset( &newSigSet, SIGUSR2 );
  }
  if ( sigismember( &oldSigSet, SIGTERM ) == 1 ) {
    sigdelset( &newSigSet, SIGTERM );
  }
  if ( sigismember( &oldSigSet, SIGINT ) == 1 ) {
    sigdelset( &newSigSet, SIGINT );
  }

  // Signale unblocken
  if ( sigprocmask( SIG_UNBLOCK, &newSigSet, 0 ) == -1 ){
    RL_LOGERROR( "sigprocmask --> errno = " << errno << " - " << strerror( errno ) );
  }
}


void RL_PP::checkPendingSignals()
{
  // Signale ueberpruefen
  sigpending( &pendingSignals );

  // wartende signale abarbeiten
  while ( (sigismember( &pendingSignals, SIGUSR1 ) == 1) ||
          (sigismember( &pendingSignals, SIGUSR2 ) == 1) ||
          (sigismember( &pendingSignals, SIGTERM ) == 1) ||
          (sigismember( &pendingSignals, SIGINT ) == 1) )
  {
    int sig = 0;
    if ( sigwait( &pendingSignals, &sig ) != 0 ) {
      RL_LOGERROR( "errno = " << errno << " --> " << strerror( errno ) );
    }
    else {
      switch( sig ) {
        case SIGUSR1:
          sigUsr1();
          break;

        case SIGUSR2:
          sigUsr2();
          break;

        case SIGTERM:
          sigTerm();
          break;

        case SIGINT:
          sigInt();
          break;

        default:
          RL_LOGERROR( "Unknown signal = " << sig );
      };
      sigdelset( &pendingSignals, sig );
    }
  }
}


void RL_PP::sigUsr1()
{
  // Konfiguration laden
  RL_LOGDATA( "SIGUSR1 received --> load configuration" );
  loadConf();
}


void RL_PP::sigUsr2()
{
  // Dummy
  RL_LOGDATA( "SIGUSR2 received --> nothing implemented here!" );
}


void RL_PP::sigTerm()
{
  RL_LOGDATA( "SIGTERM received --> stop learning (after episode)" );
  sigTermRcvd = true;
}


void RL_PP::sigInt()
{
  RL_LOGDATA( "SIGINT received --> stop learning (after step)" );
  sigIntRcvd = true;
}
