#ifndef __RL_PP_H__
#define __RL_PP_H__

#include <string>


class RL_PP
{
  public:
    enum TRACE_TYPE
    {
      NONE,
      REPLACING,
      ACCUMULATING
    };

    enum ACTION_SELECTION_TYPE
    {
      GREEDY,
      EPSILON_GREEDY,
      SOFTMAX
    };


    /**
    * Konstruktor
    * @param fileName - Filename der Konfigurationsdatei
    */
    RL_PP( const std::string& fileName );

    /**
    * Destruktor
    */
    virtual ~RL_PP();

    /**
    * Die "Lernschleife"
    */
    virtual void controlLoop();

    /**
    * Abspeichern der Konfiguration
    */
    virtual void saveConf() const;

    /**
    * Laden der Konfiguration
    */
    virtual void loadConf();



  protected:
    /**
    * Initialisieren einer Episode
    */
    virtual void initEpisode() = 0;

    /**
    * Ein Zeitschritt der Episode.
    */
    virtual void step() = 0;

    /**
    * Beenden einer Episode
    */
    virtual void finishEpisode();

    /**
    * Setzt den Startzustand der Episode.
    */
    virtual void setInitialState() = 0;

    /**
    * Aktion ausfuehren.
    */
    virtual void takeAction() = 0;

    /**
    * (neuen) Zustand ermitteln
    * @return true - Episode geht weiter, false - Episode ist vorbei
    */
    virtual const bool observeState() = 0;

    /**
    * Bestimmt den Reward.
    * @return der Reward
    */
    virtual const double calculateReward() = 0;

    /**
    * Ueberprueft ob die ValueFunctionPtr != 0 sind.
    * @return ValueFunctionPtr != 0
    */
    virtual const bool checkValueFunction() const = 0;

    /**
    * Ermittelt ob eine zufaellige Aktion gewaehlt werden soll
    * @return true - x aus [0,1) < EPSILON, false - sonst
    */
    virtual const bool randomAction() const;

    /**
    * Blocken von SIGTERM, SIGINT und SIGUSR1.
    * Diese Signale koennen dann mit sigpending abgefragt werden
    * und entsprechende Aktionen ausgefuehrt werden.
    * SIGTERM - verlassen der Controlloop nach Beendigung der Episode
    * SIGINT  - verlassen der Controlloop nach Beendigung des Steps
    * SIGUSR1 - Laden des Conf-files
    * SIGUSR2 - Frei fuer den Nutzer verfuegbar
    */
    virtual void installSignalHandler();

    /**
    * Die geblockten Signale werden (falls sie nicht auch schon vorher geblockt wurden)
    * nicht mehr geblockt.
    */
    virtual void removeSignalHandler();

    /**
    * Ueberprueft ob es wartende Signale gibt.
    */
    virtual void checkPendingSignals();

    /**
    * Wird gerufen wenn ein SIGUSR1 empfangen wird.
    */
    virtual void sigUsr1();

    /**
    * Wird gerufen wenn ein SIGUSR2 empfangen wird.
    */
    virtual void sigUsr2();

    /**
    * Wird gerufen wenn ein SIGTERM empfangen wird.
    */
    virtual void sigTerm();

    /**
    * Wird gerufen wenn ein SIGINT empfangen wird.
    */
    virtual void sigInt();


    /** Lernparameter */
    /** Lernrate */
    double ALPHA;
    /** TraceDecay */
    double LAMBDA;
    /** Discountrate */
    double GAMMA;
    /** Falls x aux [0,1) < Epsilon --> zufaellige Aktion waehlen */
    double EPSILON;
    /** Art der eligibility-traces */
    TRACE_TYPE TRACES;
    /** Dieser Wert wird auf die eligibility-traces addiert. */
    double TRACE_VALUE;
    /** Art der Aktionsauswahl */
    ACTION_SELECTION_TYPE ACTION_SELECTION;
    /** Temperature fuer Softmax (Boltzmannverteilung) */
    double TEMPERATURE;

    /** Solange gelernt werden soll --> true */
    bool continueLearning;
    /** Solange die Episode noch nicht vorbei ist --> true */
    bool continueEpisode;

    /** Gibt an ob ein SIGTERM empfangen wurde. */
    bool sigTermRcvd;
    /** Gibt an ob ein SIGINT empfangen wurde. */
    bool sigIntRcvd;

    /** Anzahl der Episoden */
    int episodes;
    /** Anzahl der Steps innerhalb einer Episode */
    int steps;
  
    /** Der Wert des aktuellen Zustands bzw. des aktiven Aktionsindex im aktuellen Zustand. */
    double value;

    /** Name der Konfigurationsdatei */
    std::string confFileName;

    /** Die bisher geblockten Signale. */
    sigset_t oldSigSet;
    /** Die zu blockenden Signale. */
    sigset_t newSigSet;
    /** Wartende Signale die geblockt wurden. */
    sigset_t pendingSignals;
};

#endif
