#ifndef __STATE_H__
#define __STATE_H__

#include "Log.h"
#include "VarValues.h"
#include "VarValuesDouble.h"
#include "VarValuesInt.h"
#include <vector>


class State
{
  public:
    enum TYPE
    {
      UNDEF,
      INT,
      DOUBLE
    };

    /**
    * Konstruktor
    */
    State();

    /**
    * Konstruktor zum Laden
    * @param file - filehandle
    */
    State( FILE* file );

    /**
    * Destruktor
    */
    ~State();

    /**
    * Anmelden einer Variablen
    * @param name - Name der Variablen
    * @param minValue - kleinster zulaessiger Wert der Variablen
    * @param maxValue - groesster zulaessiger Wert der Variablen
    * @return >= 0 - Id der Variablen,
    *         UINT_MAX bei Fehler
    */
    const unsigned int registerVariable( const std::string& name, const double minValue, const double maxValue );

    /**
    * Anmelden einer Variablen
    * @param name - Name der Variablen
    * @param minValue - kleinster zulaessiger Wert der Variablen
    * @param maxValue - groesster zulaessiger Wert der Variablen
    * @return >= 0 - Id der Variablen,
    *         UINT_MAX bei Fehler
    */
    const unsigned int registerVariable( const std::string& name, const int minValue, const int maxValue );

    /**
    * Liefert die Id der Variablen mit Name name
    * @param name - Name der Variablen
    * @return die VariablenId zur Variablen mit Namen name (UINT_MAX bei Fehler)
    */
    const unsigned int getVariableId( const std::string& name ) const;

    /**
    * Abspeichern des Zustands
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    const int save( FILE* file ) const;

    /**
    * Laden des Zustands
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    const int load( FILE* file );

    /**
    * Holt den Wert einer Variablen.
    * @param varId - Id der Variablen, deren Wert geholt werden soll
    * @return Wert der Variablen
    */
    inline const double getValueDbl( const unsigned int varId ) const;

    /**
    * Holt den kleinsten zulaessigen Wert einer Variablen.
    * @param varId - Id der Variablen, deren Wert geholt werden soll
    * @return kleinster zulaessiger Wert der Variablen
    */
    inline const double getMinValueDbl( const unsigned int varId ) const;

    /**
    * Holt den groessten zulassigen Wert einer Variablen.
    * @param varId - Id der Variablen, deren Wert geholt werden soll
    * @return groesster zulaessiger Wert der Variablen
    */
    inline const double getMaxValueDbl( const unsigned int varId ) const;

    /**
    * Holt den Wert einer Variablen.
    * @param varId - Id der Variablen, deren Wert geholt werden soll
    * @return Wert der Variablen
    */
    inline const int getValueInt( const unsigned int varId ) const;

    /**
    * Holt den kleinsten zulaessigen Wert einer Variablen.
    * @param varId - Id der Variablen, deren Wert geholt werden soll
    * @return kleinster zulaessiger Wert der Variablen
    */
    inline const int getMinValueInt( const unsigned int varId ) const;

    /**
    * Holt den groessten zulassigen Wert einer Variablen.
    * @param varId - Id der Variablen, deren Wert geholt werden soll
    * @return groesster zulaessiger Wert der Variablen
    */
    inline const int getMaxValueInt( const unsigned int varId ) const;

    /**
    * Setzt den Wert einer Variablen.
    * @param varId - Id der Variablen, deren Wert gesetzt werden soll
    * @param value - neuer Wert der Variablen
    */
    inline void setValueDbl( const unsigned int varId, double value );

    /**
    * Setzt den kleinsten zulaessigen Wert einer Variablen.
    * @param varId - Id der Variablen, deren Wert gesetzt werden soll
    * @param value - neuer kleinster zulaessiger Wert der Variablen
    */
    inline void setMinValueDbl( const unsigned int varId, const double minValue );

    /**
    * Setzt den Wert groessten zulaessigen Wert einer Variablen.
    * @param varId - Id der Variablen, deren Wert gesetzt werden soll
    * @param maxValue - neuer groesster zulaessiger Wert der Variablen
    */
    inline void setMaxValueDbl( const unsigned int varId, const double maxValue );

    /**
    * Setzt den Wert einer Variablen.
    * @param varId - Id der Variablen, deren Wert gesetzt werden soll
    * @param value - neuer Wert der Variablen
    */
    inline void setValueInt( const unsigned int varId, int value );

    /**
    * Setzt den kleinsten zulaessigen Wert einer Variablen.
    * @param varId - Id der Variablen, deren Wert gesetzt werden soll
    * @param value - neuer kleinster zulaessiger Wert der Variablen
    */
    inline void setMinValueInt( const unsigned int varId, const int minValue );

    /**
    * Setzt den Wert groessten zulaessigen Wert einer Variablen.
    * @param varId - Id der Variablen, deren Wert gesetzt werden soll
    * @param maxValue - neuer groesster zulaessiger Wert der Variablen
    */
    inline void setMaxValueInt( const unsigned int varId, const int maxValue );

    /**
    * Gibt die Anzahl der Variablen zurueck.
    * @return VARIABLES
    */
    inline const unsigned int getNumberOfVariables() const;


  protected:
    // der Typ
    TYPE varType;

    /** Enthaelt die Werte der registrierten Variablen. */
    std::vector< VarValues* > vars;
};


inline const double State::getValueDbl( const unsigned int varId ) const
{
  if ( varType != DOUBLE ) {
    RL_LOGERROR( "Not a Double-State!" );
    return 0.0;
  }

  if ( (varId < 0) || (varId >= vars.size()) ) {
    RL_LOGERROR( "VarId out of bounds --> 0 <= " << varId << " < " << vars.size() );
    return 0.0;
  }

  return static_cast<VarValuesDouble*>(vars[ varId ])->getValue();
}

inline const double State::getMinValueDbl( const unsigned int varId ) const
{
  if ( varType != DOUBLE ) {
    RL_LOGERROR( "Not a Double-State!" );
    return 0.0;
  }

  if ( (varId < 0) || (varId >= vars.size()) ) {
    RL_LOGERROR( "VarId out of bounds --> 0 <= " << varId << " < " << vars.size() );
    return 0.0;
  }

  return static_cast<VarValuesDouble*>(vars[ varId ])->getMinValue();
}

inline const double State::getMaxValueDbl( const unsigned int varId ) const
{
  if ( varType != DOUBLE ) {
    RL_LOGERROR( "Not a Double-State!" );
    return 0.0;
  }

  if ( (varId < 0) || (varId >= vars.size()) ) {
    RL_LOGERROR( "VarId out of bounds --> 0 <= " << varId << " < " << vars.size() );
    return 0.0;
  }

  return static_cast<VarValuesDouble*>(vars[ varId ])->getMaxValue();
}

inline const int State::getValueInt( const unsigned int varId ) const
{
  if ( varType != INT ) {
    RL_LOGERROR( "Not an Int-State!" );
    return 0;
  }

  if ( (varId < 0) || (varId >= vars.size()) ) {
    RL_LOGERROR( "VarId out of bounds --> 0 <= " << varId << " < " << vars.size() );
    return 0;
  }

  return static_cast<VarValuesInt*>(vars[ varId ])->getValue();
}

inline const int State::getMinValueInt( const unsigned int varId ) const
{
  if ( varType != INT ) {
    RL_LOGERROR( "Not an Int-State!" );
    return 0;
  }

  if ( (varId < 0) || (varId >= vars.size()) ) {
    RL_LOGERROR( "VarId out of bounds --> 0 <= " << varId << " < " << vars.size() );
    return 0;
  }

  return static_cast<VarValuesInt*>(vars[ varId ])->getMinValue();
}

inline const int State::getMaxValueInt( const unsigned int varId ) const
{
  if ( varType != INT ) {
    RL_LOGERROR( "Not an Int-State!" );
    return 0;
  }

  if ( (varId < 0) || (varId >= vars.size()) ) {
    RL_LOGERROR( "VarId out of bounds --> 0 <= " << varId << " < " << vars.size() );
    return 0;
  }

  return static_cast<VarValuesInt*>(vars[ varId ])->getMaxValue();
}

inline void State::setValueDbl( const unsigned int varId, double value )
{
  if ( varType != DOUBLE ) {
    RL_LOGERROR( "Not a Double-State!" );
    return;
  }

  VarValuesDouble* varPtr = static_cast<VarValuesDouble*>(vars[ varId ]);

  if ( (varId < 0) || (varId >= vars.size()) ) {
    RL_LOGERROR( "VarId out of bounds --> 0 <= " << varId << " < " << vars.size() );
    return;
  }

  if ( value < varPtr->getMinValue() ) {
    RL_LOGWARNING( "value(" << value << ") < minValue(" << varPtr->getMinValue() << ")" );
    value = varPtr->getMinValue();
  }

  if ( value > varPtr->getMaxValue() ) {
    RL_LOGWARNING( "value(" << value << ") > maxValue(" << varPtr->getMaxValue() << ")" );
    value = varPtr->getMaxValue();
  }

  varPtr->setValue( value );
}

inline void State::setMinValueDbl( const unsigned int varId, const double minValue )
{
  if ( varType != DOUBLE ) {
    RL_LOGERROR( "Not a Double-State!" );
    return;
  }

  VarValuesDouble* varPtr = static_cast<VarValuesDouble*>(vars[ varId ]);

  if ( (varId < 0) || (varId >= vars.size()) ) {
    RL_LOGERROR( "VarId out of bounds --> 0 <= " << varId << " < " << vars.size() );
    return;
  }

  if ( minValue > varPtr->getMaxValue() ) {
    RL_LOGWARNING( "minValue( " << minValue << ") > maxValue( " << varPtr->getMaxValue() << ")" );
    varPtr->setMinValue( varPtr->getMaxValue() );
    varPtr->setMaxValue( minValue );
    return;
  }

  varPtr->setMinValue( minValue );
}

inline void State::setMaxValueDbl( const unsigned int varId, const double maxValue )
{
  if ( varType != DOUBLE ) {
    RL_LOGERROR( "Not a Double-State!" );
    return;
  }

  VarValuesDouble* varPtr = static_cast<VarValuesDouble*>(vars[ varId ]);

  if ( (varId < 0) || (varId >= vars.size()) ) {
    RL_LOGERROR( "VarId out of bounds --> 0 <= " << varId << " < " << vars.size() );
    return;
  }

  if ( maxValue < varPtr->getMinValue() ) {
    RL_LOGWARNING( "maxValue( " << maxValue << ") < minValue( " << varPtr->getMinValue() << ")" );
    varPtr->setMaxValue( varPtr->getMinValue() );
    varPtr->setMinValue( maxValue );
    return;
  }

  varPtr->setMaxValue( maxValue );
}

inline void State::setValueInt( const unsigned int varId, int value )
{
  if ( varType != INT ) {
    RL_LOGERROR( "Not an Int-State!" );
    return;
  }

  VarValuesInt* varPtr = static_cast<VarValuesInt*>(vars[ varId ]);

  if ( (varId < 0) || (varId >= vars.size()) ) {
    RL_LOGERROR( "VarId out of bounds --> 0 <= " << varId << " < " << vars.size() );
    return;
  }

  if ( value < varPtr->getMinValue() ) {
    RL_LOGWARNING( "value(" << value << ") < minValue(" << varPtr->getMinValue() << ")" );
    value = varPtr->getMinValue();
  }

  if ( value > varPtr->getMaxValue() ) {
    RL_LOGWARNING( "value(" << value << ") > maxValue(" << varPtr->getMaxValue() << ")" );
    value = varPtr->getMaxValue();
  }

  varPtr->setValue( value );
}

inline void State::setMinValueInt( const unsigned int varId, const int minValue )
{
  if ( varType != INT ) {
    RL_LOGERROR( "Not an Int-State!" );
    return;
  }

  VarValuesInt* varPtr = static_cast<VarValuesInt*>(vars[ varId ]);

  if ( (varId < 0) || (varId >= vars.size()) ) {
    RL_LOGERROR( "VarId out of bounds --> 0 <= " << varId << " < " << vars.size() );
    return;
  }

  if ( minValue > varPtr->getMaxValue() ) {
    RL_LOGWARNING( "minValue( " << minValue << ") > maxValue( " << varPtr->getMaxValue() << ")" );
    varPtr->setMinValue( varPtr->getMaxValue() );
    varPtr->setMaxValue( minValue );
    return;
  }

  varPtr->setMinValue( minValue );
}

inline void State::setMaxValueInt( const unsigned int varId, const int maxValue )
{
  if ( varType != INT ) {
    RL_LOGERROR( "Not an Int-State!" );
    return;
  }

  VarValuesInt* varPtr = static_cast<VarValuesInt*>(vars[ varId ]);

  if ( (varId < 0) || (varId >= vars.size()) ) {
    RL_LOGERROR( "VarId out of bounds --> 0 <= " << varId << " < " << vars.size() );
    return;
  }

  if ( maxValue < varPtr->getMinValue() ) {
    RL_LOGWARNING( "maxValue( " << maxValue << ") < minValue( " << varPtr->getMinValue() << ")" );
    varPtr->setMaxValue( varPtr->getMinValue() );
    varPtr->setMinValue( maxValue );
    return;
  }

  varPtr->setMaxValue( maxValue );
}

inline const unsigned int State::getNumberOfVariables() const
{
  return vars.size();
}

#endif
