#include "StateLookupTable.h"
#include "Log.h"
#include "State.h"

#include <cmath>
#include <cfloat>


StateLookupTable::StateLookupTable( State* aStatePtr )
  : StateValueFunction( aStatePtr ),
    LookupTable( aStatePtr )
{
  // den Funktionspointer initialisieren
  value = 0.0;
  initWeightsFunctionPtr = &getValue;

  // Speicher holen
  LookupTable::allocateMemory();

  // indexOffset
  for ( unsigned int v = 0; v < StateValueFunction::statePtr->getNumberOfVariables(); ++v ) {
    indexOffset[v] = 1;
    for ( unsigned int i = v + 1; i < StateValueFunction::statePtr->getNumberOfVariables(); ++i ) {
      // +1 --> damit lower und upper Bound im Bereich sind
      indexOffset[v] *= (StateValueFunction::statePtr->getMaxValueInt( i ) - StateValueFunction::statePtr->getMinValueInt( i ) + 1);
    }
  }

  // Anzahl der Gewichte bestimmen
  ARRAY_SIZE = 1;
  for ( unsigned int v = 0; v < StateValueFunction::statePtr->getNumberOfVariables(); ++v ) {
    // +1 --> damit lower und upper Bound im Bereich sind
    ARRAY_SIZE *= (StateValueFunction::statePtr->getMaxValueInt( v ) - StateValueFunction::statePtr->getMinValueInt( v ) + 1);
  }

  // Gewichtsvektor anlegen
  weights = LookupTable::allocateMemoryWeights();

  // Gewichtsvektor initialisieren
  initWeights( initWeightsFunctionPtr );
}


StateLookupTable::StateLookupTable( const std::string& fileName )
  : StateValueFunction( 0 ),
    LookupTable( 0 )
{
  // den Funktionspointer initialisieren
  value = 0.0;
  initWeightsFunctionPtr = &getValue;

  // Laden
  StateValueFunction::load( fileName );
}


StateLookupTable::~StateLookupTable()
{
}


void StateLookupTable::initWeights( const double (*initFunctionPtr)() )
{
  LookupTable::initWeights( initFunctionPtr );
}


void StateLookupTable::initEligibilityTraces()
{
  LookupTable::initEligibilityTraces();
}


void StateLookupTable::updateWeights( const double ALPHA, const double DELTA )
{
  LookupTable::updateWeights( ALPHA, DELTA );
}


void StateLookupTable::updateEligibilityTraces( const double GAMMA, const double LAMBDA )
{
  LookupTable::updateEligibilityTraces( GAMMA, LAMBDA );
}


void StateLookupTable::accumulatingTraces( const double VALUE )
{
  if ( traceStorage == Types::ARRAY ) {
    if ( eTraces == 0 ) {
      RL_LOGERROR( "traceStorage == ARRAY and eTraces == 0"
                   << " --> Set traceStorage to MAP or call initEligibilityTraces before calling updateWeights!" );
      return;
    }

    if ( activeIndex != -1 ) {
      eTraces[ activeIndex ] += VALUE;
    }
  }
  else if ( traceStorage == Types::MAP ) {
    std::map< int, double >::iterator i;

    if ( activeIndex != -1 ) {
      i = eTracesMap.find( activeIndex );
      if ( i == eTracesMap.end() ) {
        // Key existiert noch nicht --> anlegen
        eTracesMap[ activeIndex ] = VALUE;
      }
      else {
        // Key existiert schon --> Wert aufaddieren
        i->second += VALUE;
      }
    }
  }
  else {
    RL_LOGERROR( "Unknown traceStorage = " << traceStorage );
  }
}


void StateLookupTable::replacingTraces( const double VALUE )
{
  if ( traceStorage == Types::ARRAY ) {
    if ( eTraces == 0 ) {
      RL_LOGERROR( "traceStorage == ARRAY and eTraces == 0"
                   << " --> Set traceStorage to MAP or call initEligibilityTraces before calling updateWeights!" );
      return;
    }

    if ( activeIndex != -1 ) {
      eTraces[ activeIndex ] = VALUE;
    }
  }
  else if ( traceStorage == Types::MAP ) {
    if ( activeIndex != -1 ) {
      eTracesMap[ activeIndex ] = VALUE;
    }
  }
  else {
    RL_LOGERROR( "Unknown traceStorage = " << traceStorage );
  }
}


void StateLookupTable::storeActiveIndex()
{
  storedActiveIndex = activeIndex;
}


const double StateLookupTable::getStateValue() const
{
  if ( weights != 0 ) {
    if ( activeIndex != -1 ) {
      return weights[ activeIndex ];
    }
    else {
      return 0.0;
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
    return DBL_MAX;
  }
}


void StateLookupTable::storeEpisodeState()
{
  if ( activeIndex != -1 ) {
    episodeIndexHistorySet.insert( activeIndex );
  }
}


void StateLookupTable::storeState()
{
  storeActiveIndex();
  storeEpisodeState();
}


void StateLookupTable::setState()
{
  LookupTable::setState();
}


void StateLookupTable::initEpisodeStateHistory()
{
  LookupTable::initEpisodeStateHistory();
}


void StateLookupTable::valueEpisode( const double value )
{
  LookupTable::valueEpisode( value );
}


void StateLookupTable::checkWeightLimit()
{
  LookupTable::checkWeightLimit( minWeightLimit, maxWeightLimit );
}


const double StateLookupTable::getMinWeight() const
{
  return LookupTable::getMinWeight();
}


const double StateLookupTable::getAvgWeight( const bool abs ) const
{
  return LookupTable::getAvgWeight( abs );
}


const double StateLookupTable::getMaxWeight() const
{
  return LookupTable::getMaxWeight();
}


const int StateLookupTable::save( FILE* file ) const
{
  return LookupTable::save( file );
}


const int StateLookupTable::load( FILE* file )
{
  // jetzt ist der StatePtr bereits geladen --> auch im LookupTable setzen
  LookupTable::setStatePtr( ValueFunction::statePtr );
  return LookupTable::load( file );
}
