#ifndef __STATE_LOOKUP_TABLE_H__
#define __STATE_LOOKUP_TABLE_H__

#include "StateValueFunction.h"
#include "LookupTable.h"


class StateLookupTable : public StateValueFunction,
                         public LookupTable
{
  public:
    /**
    * Konstruktor
    * @param aStatePtr - Pointer auf den Zustand
    */
    StateLookupTable( State* aStatePtr );

    /**
    * Konstruktor zum Laden
    * @param fileName - Name der zu ladenden Datei
    */
    StateLookupTable( const std::string& fileName );

    /**
    * Destruktor
    */
    virtual ~StateLookupTable();

    /**
    * Initialisiert die Gewichte
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    * @param initFunctionPtr - Pointer auf die Funktion die zum Initialisieren bentzt wird
    */
    virtual void initWeights( const double (*initFunctionPtr)() );

    /**
    * Initialisieren der EligibilityTraces mit 0.0
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    */
    virtual void initEligibilityTraces();

    /**
    * Updaten der Gewichte
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    * @param ALPHA - die Lernrate
    * @param DELTA - die Aenderung des Gewichts
    */
    virtual void updateWeights( const double ALPHA, const double DELTA );

    /**
    * Updaten der EligibilityTraces
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    * @param GAMMA - Discountrate
    * @param LAMBDA - TraceDecay
    */
    virtual void updateEligibilityTraces( const double GAMMA, const double LAMBDA );

    /**
    * Die Traces der aktivierten Features werden erhoeht.
    * @param VALUE - der Wert um den erhoeht wird
    */
    virtual void accumulatingTraces( const double VALUE );

    /**
    * Die Traces der aktivierten Features werden gesetzt.
    * @param VALUE - der Wert auf den gesetzt wird
    */
    virtual void replacingTraces( const double VALUE );

    /**
    * Falls keine EligibilityTraces verwendet werden, muss der Zustand gespeichert
    * werden, damit man weiss welche Gewichte geupdatet werden muessen.
    * --> den aktivierten Index speichern
    */
    virtual void storeActiveIndex();

    /**
    * Liefert den Wert des Zustands mit dem zuvor aktivierten Index.
    * @return Wert des Zustands
    */
    virtual const double getStateValue() const;

    /**
    * Speichert den Zustand in der Historie dieser Episode ab.
    * Dazu werden alle Indizees der aktiven Gewichte gespeichert.
    */
    virtual void storeEpisodeState();

    /**
    * Falls keine EligibilityTraces verwendet werden, muss der Zustand gespeichert
    * werden, damit man weiss welche Gewichte geupdatet werden muessen.
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    */
    virtual void storeState();

    /**
    * Setzt den Zustand --> aktiviert die Tilings
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    */
    virtual void setState();

    /**
    * Initialisiert die episodeStateHistory.
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    */
    virtual void initEpisodeStateHistory();

    /**
    * Bewertet die gesamte Episode.
    * --> dafuer muss episodeStateHistory auf true gesetzt sein
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    * @param value - dieser Wert wird auf alle Gewichte der Episode gerechnet
    */
    virtual void valueEpisode( const double value );

    /**
    * Ueberprueft ob die Gewichte in den Grenzen sind.
    * minWeight <= weight <= maxWeight
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    */
    virtual void checkWeightLimit();

    /**
    * Liefert das minimale Gewicht zurueck.
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    * @return das minimale Gewicht
    */
    virtual const double getMinWeight() const;

    /**
    * Liefert das durchschnittliche Gewicht zurueck.
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    * @param abs - wenn true wird der Durchschnitt ueber den absoluten Betrag der Gewichte gebildet
    * @return das durchschnittliche Gewicht
    */
    virtual const double getAvgWeight( const bool abs ) const;

    /**
    * Liefert das maximale Gewicht zurueck.
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    * @return das maximale Gewicht
    */
    virtual const double getMaxWeight() const;


  protected:
    /**
    * Abspeichern
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int save( FILE* file ) const;

    /**
    * Laden
    * --> Diese Implementation ruft nur die Implementation aus LookupTable auf!
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int load( FILE* file );
};

#endif
