#include "TileCoding.h"
#include "Log.h"
#include "Tilings.h"

#include <cerrno>
#include <cfloat>


TileCoding::TileCoding()
{
  // Anzahl der Tilings bevor Tilings registered werden natuerlich 0
  numberOfTilings = 0;
}


TileCoding::~TileCoding()
{
  freeMemory();
}


void TileCoding::initWeights( const double (*initFunctionPtr)() )
{
  for ( std::list< Tilings* >::iterator t = tilingPtrList.begin();
        t != tilingPtrList.end(); ++t )
  {
    (*t)->initWeights( initFunctionPtr );
  }
}


void TileCoding::initEligibilityTraces()
{
  for ( std::list< Tilings* >::iterator t = tilingPtrList.begin();
        t != tilingPtrList.end(); ++t )
  {
    (*t)->initEligibilityTraces();
  }
}


void TileCoding::updateWeights( const double ALPHA, const double DELTA )
{
  for ( std::list< Tilings* >::iterator t = tilingPtrList.begin();
        t != tilingPtrList.end(); ++t )
  {
    (*t)->updateWeights( (ALPHA / numberOfTilings), DELTA );
  }
}


void TileCoding::updateEligibilityTraces( const double GAMMA, const double LAMBDA )
{
  for ( std::list< Tilings* >::iterator t = tilingPtrList.begin();
        t != tilingPtrList.end(); ++t )
  {
    (*t)->updateEligibilityTraces( GAMMA, LAMBDA );
  }
}


void TileCoding::setTraceStorage( const Types::TRACE_STORAGE newTraceStorage )
{
  for ( std::list< Tilings* >::iterator t = tilingPtrList.begin();
        t != tilingPtrList.end(); ++t )
  {
    (*t)->setTraceStorage( newTraceStorage );
  }
}


const double TileCoding::getMinWeight() const
{
  double min = DBL_MAX;
  double value;

  for ( std::list< Tilings* >::const_iterator t = tilingPtrList.begin();
        t != tilingPtrList.end(); ++t )
  {
    value = (*t)->getMinWeight();
    if ( value < min ) {
      min = value;
    }
  }

  return min;
}


const double TileCoding::getAvgWeight( const bool abs ) const
{
  double sum = 0.0;
  int weightCount = 0;

  for ( std::list< Tilings* >::const_iterator t = tilingPtrList.begin();
        t != tilingPtrList.end(); ++t )
  {
    sum += (*t)->getWeightSum( abs, weightCount );
  }

  return (sum / weightCount);
}


const double TileCoding::getMaxWeight() const
{
  double max = -1.0 * DBL_MAX;
  double value;

  for ( std::list< Tilings* >::const_iterator t = tilingPtrList.begin();
        t != tilingPtrList.end(); ++t )
  {
    value = (*t)->getMaxWeight();
    if ( value > max ) {
      max = value;
    }
  }

  return max;
}


void TileCoding::setState()
{
  for ( std::list< Tilings* >::const_iterator t = tilingPtrList.begin();
        t != tilingPtrList.end(); ++t )
  {
    (*t)->activateTiles();
  }
}


void TileCoding::initEpisodeStateHistory()
{
  for ( std::list< Tilings* >::const_iterator t = tilingPtrList.begin();
        t != tilingPtrList.end(); ++t )
  {
    (*t)->initEpisodeStateHistory();
  }
}


void TileCoding::valueEpisode( const double value )
{
  for ( std::list< Tilings* >::const_iterator t = tilingPtrList.begin();
        t != tilingPtrList.end(); ++t )
  {
    (*t)->valueEpisode( value );
  }
}


const int TileCoding::save( FILE* file ) const
{
  // Anzahl der Tilings
  unsigned int listSize = tilingPtrList.size();
  if ( fwrite( &listSize, sizeof( unsigned int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << " --> " << strerror( errno ) );
    return -1;
  }

  // Tilings
  for ( std::list< Tilings* >::const_iterator t = tilingPtrList.begin();
        t != tilingPtrList.end(); ++t )
  {
    int returnValue = (*t)->save( file );
    if ( returnValue != 0 ) {
      RL_LOGERROR( "Tiling::save returned " << returnValue );
      return -1;
    }
  }

  return 0;
}


void TileCoding::freeMemory()
{
  for ( std::list< Tilings* >::iterator t = tilingPtrList.begin();
        t != tilingPtrList.end(); ++t )
  {
    delete (*t);
  }
  tilingPtrList.clear();
  numberOfTilings = 0;
}
