#include "Tilings.h"

#include "Log.h"
#include "State.h"
#include "Utility.h"

#include <cerrno>
#include <cfloat>
#include <cmath>
#include <iostream>



Tilings::Tilings( const int numberOfTilings, const std::list< Types::TILINGS_VAR_STRUCT >& varList )
{
  // Anzahl der Tilings
  TILINGS = numberOfTilings;

  // Anzahl der Variablen/Dimensionen
  DIMENSIONS = varList.size();

  // Wie werden die Traces gespeichert Array oder Map? --> default MAP
  traceStorage = Types::MAP;

  // Pointer auf den Zustand --> wird spaeter bei create gesetzt
  statePtr = 0;

  // Speicher holen --> kann man jetzt schon machen, da nur TILINGS und DIMENSIONS benoetigt werden
  allocateMemory();

  // storedActiveTiles
  storedActiveIndices = 0;

  // vars sichern
  int d = 0;
  for ( std::list< Types::TILINGS_VAR_STRUCT >::const_iterator v = varList.begin();
        v != varList.end(); ++v  )
  {
    vars[d] = *v;
    ++d;
  }

  // weights --> werden erst bei create angelegt
  weights = 0;
  // EligibilityTraces werden erst angelegt wenn sie benoetigt werden
  eTraces = 0;
}


Tilings::Tilings( State* aStatePtr, FILE* file )
{
  // Wie werden die Traces gespeichert Array oder Map? --> default ARRAY
  traceStorage = Types::MAP;
  // Pointer auf den Zustand
  statePtr = aStatePtr;
  // vars
  vars = 0;
  // tiles
  tiles = 0;
  // tileWidth
  tileWidth = 0;
  // indexOffset
  indexOffset = 0;
  // activeTiles
  activeTiles = 0;
  // storedActiveTiles
  storedActiveIndices = 0;
  // tileOffset
  tileOffset = 0;
  // weights
  weights = 0;
  // EligibilityTraces werden erst angelegt wenn sie benoetigt werden
  eTraces = 0;

  // Laden
  load( file );
}


Tilings::~Tilings()
{
  freeMemory();
}


void Tilings::initWeights( const double (*initFunctionPtr)() )
{
  if ( weights != 0 ) {
    for ( int w = 0; w < ARRAY_SIZE; ++w ) {
      weights[w] = (*initFunctionPtr)();
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }
}


void Tilings::initEligibilityTraces()
{
  if ( traceStorage == Types::ARRAY ) {
    if ( eTraces == 0 ) {
      // EligibilityTraces sind noch nicht angelegt --> Speicher holen
      eTraces = allocateMemoryWeights();
    }

    for ( int e = 0; e < ARRAY_SIZE; ++e ) {
      eTraces[e] = 0.0;
    }
  }
  else if ( traceStorage == Types::MAP ) {
    clearTraceMap();
  }
  else {
    RL_LOGERROR( "Unknown traceStorage = " << traceStorage );
  }
}


void Tilings::updateWeights( const double ALPHA, const double DELTA )
{
  if ( weights != 0 ) {
    const double FACTOR = ALPHA * DELTA;

    if ( traceStorage == Types::ARRAY ) {
      if ( eTraces != 0 ) {
        for ( int w = 0; w < ARRAY_SIZE; ++w ) {
          weights[w] += FACTOR * eTraces[w];
        }
      }
      else if ( storedActiveIndices != 0 ) {
        for ( int t = 0; t < TILINGS; ++t ) {
          if ( storedActiveIndices[t] != -1 ) {
            weights[ storedActiveIndices[t] ] += FACTOR;
          }
        }
      }
      else {
        RL_LOGERROR( "traceStorage == ARRAY and eTraces == 0 and storedActiveIndices == 0"
                     << " --> Set traceStorage to MAP"
                     << " or call initEligibilityTraces or storeState before calling updateWeights!" );
      }
    }
    else if ( traceStorage == Types::MAP ) {
      if ( eTracesMap.empty() == false ) {
        for ( std::map< int, double >::const_iterator i = eTracesMap.begin();
            i != eTracesMap.end(); ++i )
        {
          weights[ i->first ] += FACTOR * i->second;
        }
      }
      else if ( storedActiveIndices != 0 ) {
        for ( int t = 0; t < TILINGS; ++t ) {
          if ( storedActiveIndices[t] != -1 ) {
            weights[ storedActiveIndices[t] ] += FACTOR;
          }
        }
      }
      else {
        RL_LOGERROR( "traceStorage == MAP and eTracesMap == empty and storedActiveIndices == 0"
                     << " --> Set traceStorage to ARRAY"
                     << " or call initEligibilityTraces or storeState before calling updateWeights!" );
      }
    }
    else {
      RL_LOGERROR( "Unknown traceStorage = " << traceStorage );
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }
}


void Tilings::updateEligibilityTraces( const double GAMMA, const double LAMBDA )
{
  const double FACTOR = GAMMA * LAMBDA;

  if ( traceStorage == Types::ARRAY ) {
    if ( eTraces == 0 ) {
      RL_LOGERROR( "traceStorage == ARRAY and eTraces == 0"
                   << " --> Set traceStorage to MAP or call initEligibilityTraces before calling updateEligibilityTraces!" );
      return;
    }

    for ( int e = 0; e < ARRAY_SIZE; ++e ) {
      eTraces[e] *= FACTOR;
    }
  }
  else if ( traceStorage == Types::MAP ) {
    for ( std::map< int, double >::iterator i = eTracesMap.begin();
          i != eTracesMap.end(); ++i )
    {
      i->second *= FACTOR;
    }
  }
  else {
    RL_LOGERROR( "Unknown traceStorage = " << traceStorage );
  }
}


void Tilings::checkWeightLimit( const double MIN_WEIGHT, const double MAX_WEIGHT )
{
  if ( weights != 0 ) {
    double* weightPtr = &weights[0];
    for ( int w = 0; w < ARRAY_SIZE; ++w ) {
      *weightPtr = utility::minMax( *weightPtr, MIN_WEIGHT, MAX_WEIGHT );
      ++weightPtr;
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }
}


const double Tilings::getMinWeight() const
{
  double min = DBL_MAX;

  if ( weights != 0 ) {
    double* weightPtr = &weights[0];
    for ( int w = 0; w < ARRAY_SIZE; ++w ) {
      if ( *weightPtr < min ) {
        min = *weightPtr;
      }
      ++weightPtr;
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }

  return min;
}


const double Tilings::getWeightSum( const bool abs, int& weightCount ) const
{
  if ( weights != 0 ) {
    double sum = 0.0;
    double* weightPtr = &weights[0];
    if ( abs ) {
      for ( int w = 0; w < ARRAY_SIZE; ++w ) {
        sum += fabs( *weightPtr );
        ++weightPtr;
      }
    }
    else {
      for ( int w = 0; w < ARRAY_SIZE; ++w ) {
        sum += *weightPtr;
        ++weightPtr;
      }
    }

    weightCount += ARRAY_SIZE;
    return sum;
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
    return 0;
  }
}


const double Tilings::getMaxWeight() const
{
  double max = -1.0 * DBL_MAX;

  if ( weights != 0 ) {
    double* weightPtr = &weights[0];
    for ( int w = 0; w < ARRAY_SIZE; ++w ) {
      if ( *weightPtr > max ) {
        max = *weightPtr;
      }
      ++weightPtr;
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
  }

  return max;
}


void Tilings::activateTiles()
{
  if ( statePtr != 0 ) {
    for ( int t = 0; t < TILINGS; ++t ) {
      int index = t;
      for ( int d = 0; d < DIMENSIONS; ++d ) {
        switch ( vars[d].tileType ) {
          case Types::CIRCLE:
            index += indexOffset[d] * static_cast< int >( floor((utility::normalize( statePtr->getValueDbl( vars[d].id ) - tileOffset[t][d], statePtr->getMinValueDbl( vars[d].id ), statePtr->getMaxValueDbl( vars[d].id ) ) - statePtr->getMinValueDbl( vars[d].id )) / tileWidth[d]) );
            break;

          case Types::CLIP:
            if ( statePtr->getValueDbl( vars[d].id ) - tileOffset[t][d] >= statePtr->getMinValueDbl( vars[d].id ) ) {
              index += indexOffset[d] * utility::minMax( static_cast< int >( floor((statePtr->getValueDbl( vars[d].id ) - statePtr->getMinValueDbl( vars[d].id ) - tileOffset[t][d]) / tileWidth[d]) ), 0, tiles[d]-1 );
            }
            else {
              index = -1;
            }
            break;

          case Types::WRAP:
            index += indexOffset[d] * static_cast< int >( floor((statePtr->getValueDbl( vars[d].id ) - statePtr->getMinValueDbl( vars[d].id ) + tileOffset[t][d]) / tileWidth[d]) );
            break;
  
          default:
            RL_LOGERROR( "Unknown Tiletype[" << vars[d].tileType
                      << "] for Variable with id = " << vars[d].id );
        }

        if ( index == -1 ) {
          break;
        }
      }
      activeTiles[t] = index;

      if ( (index >= ARRAY_SIZE) || (index < -1) ) {
        RL_LOGERROR( "index out ouf bounds --> " << index );
      }
    }
  }
  else {
    RL_LOGERROR( "statePtr == 0" );
    for ( int t = 0; t < TILINGS; ++t ) {
      activeTiles[t] = -1;
    }
  }
}


void Tilings::initEpisodeStateHistory()
{
  episodeIndexHistorySet.clear();
}


void Tilings::valueEpisode( const double value )
{
  if ( episodeIndexHistorySet.empty() == false ) {
    if ( weights != 0 ) {
      for ( std::set< int >::const_iterator i = episodeIndexHistorySet.begin();
            i != episodeIndexHistorySet.end(); ++i )
      {
        weights[ *i ] += value;
      }
    }
    else {
      RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
    }
  }
  else {
    RL_LOGERROR( "episodeIndexHistorySet is empty! "
                 << "Did you forget to set ValueFunction::episodeStateHistory to true?" );
  }
}


const int Tilings::save( FILE* file ) const
{
  // Anzahl der Tilings
  if ( fwrite( &TILINGS, sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Anzahl der Dimensionen
  if ( fwrite( &DIMENSIONS, sizeof( int ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Art der EligibilityTraces (ARRAY, MAP)
  if ( fwrite( &traceStorage, sizeof( Types::TRACE_STORAGE ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Die Daten der Variablen
  if ( fwrite( &vars[0], sizeof( Types::TILINGS_VAR_STRUCT ), DIMENSIONS, file ) != static_cast< size_t >( DIMENSIONS ) ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Groesse eines Tiles fuer jede Variable.
  if ( fwrite( &tileWidth[0], sizeof( double ), DIMENSIONS, file ) != static_cast< size_t >( DIMENSIONS ) ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Anzahl der fuer jede Variable verwendeten Tiles
  if ( fwrite( &tiles[0], sizeof( int ), DIMENSIONS, file ) != static_cast< size_t >( DIMENSIONS ) ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Indexoffset der Variablen fuer weights und eTraces
  if ( fwrite( &indexOffset[0], sizeof( int ), DIMENSIONS, file ) != static_cast< size_t >( DIMENSIONS ) ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Fuer jedes Tiling und jede Variable das Offset
  for ( int t = 0; t < TILINGS; ++t ) {
    if ( fwrite( tileOffset[t], sizeof( double ), DIMENSIONS, file ) != static_cast< size_t >( DIMENSIONS ) ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
      return -1;
    }
  }

  // Die Anzahl der Gewichte
  if ( fwrite( &ARRAY_SIZE, sizeof( int ), 1, file ) != 1 ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Die zu lernenden Werte
  if ( weights != 0 ) {
    if ( fwrite( &weights[0], sizeof( double ), ARRAY_SIZE, file ) != static_cast< size_t >( ARRAY_SIZE ) ) {
        RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
      return -1;
    }
  }
  else {
    RL_LOGERROR( "weights == 0 --> Maybe forgot to call create?" );
    return -1;
  }

  return 0;
}


const int Tilings::load( FILE* file )
{
  // falls schon Daten vorhanden sind --> Speicher freigeben und neu allozieren
  freeMemory();

  // Anzahl der Tilings
  if ( fread( &TILINGS, sizeof( int ), 1, file ) != 1 ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Anzahl der Dimensionen
  if ( fread( &DIMENSIONS, sizeof( int ), 1, file ) != 1 ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Art der EligibilityTraces (ARRAY, MAP)
  if ( fread( &traceStorage, sizeof( Types::TRACE_STORAGE ), 1, file ) != 1 ) {
    RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Speicher holen
  allocateMemory();

  // Die Daten der Variablen
  if ( fread( &vars[0], sizeof( Types::TILINGS_VAR_STRUCT ), DIMENSIONS, file ) != static_cast< size_t >( DIMENSIONS ) ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Groesse eines Tiles fuer jede Variable.
  if ( fread( &tileWidth[0], sizeof( double ), DIMENSIONS, file ) != static_cast< size_t >( DIMENSIONS ) ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Anzahl der fuer jede Variable verwendeten Tiles
  if ( fread( &tiles[0], sizeof( int ), DIMENSIONS, file ) != static_cast< size_t >( DIMENSIONS ) ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Indexoffset der Variablen fuer weights und eTraces
  if ( fread( &indexOffset[0], sizeof( int ), DIMENSIONS, file ) != static_cast< size_t >( DIMENSIONS ) ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Fuer jedes Tiling und jede Variable das Offset
  for ( int t = 0; t < TILINGS; ++t ) {
    if ( fread( tileOffset[t], sizeof( double ), DIMENSIONS, file ) != static_cast< size_t >( DIMENSIONS ) ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
      return -1;
    }
  }

  // Die Anzahl der Gewichte
  if ( fread( &ARRAY_SIZE, sizeof( int ), 1, file ) != 1 ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  // Speicher fuer weights holen
  weights = allocateMemoryWeights();

  // Die zu lernenden Werte
  if ( fread( &weights[0], sizeof( double ), ARRAY_SIZE, file ) != static_cast< size_t >( ARRAY_SIZE ) ) {
      RL_LOGERROR( "errno = " << errno << "  -->  " << strerror( errno ) );
    return -1;
  }

  return 0;
}


void Tilings::allocateMemory()
{
  // vars
  vars = new Types::TILINGS_VAR_STRUCT[ DIMENSIONS ];
  // tiles
  tiles = new int[ DIMENSIONS ];
  // tileWidth
  tileWidth = new double[ DIMENSIONS ];
  // indexOffset
  indexOffset = new int[ DIMENSIONS ];
  // activeTiles
  activeTiles = new int[ TILINGS ];
  // tileOffset
  tileOffset = new double*[ TILINGS ];
  for ( int t = 0; t < TILINGS; ++t ) {
    tileOffset[t] = new double[ DIMENSIONS ];
  }
}


double* Tilings::allocateMemoryWeights()
{
  // Weightvektor anlegen
  return (new double[ ARRAY_SIZE ]);
}


void Tilings::freeMemory()
{
  if ( vars != 0 ) {
    delete[] vars;
    vars = 0;
  }
  if ( tileWidth != 0 ) {
    delete[] tileWidth;
    tileWidth = 0;
  }
  if ( tiles != 0 ) {
    delete[] tiles;
    tiles = 0;
  }
  if ( indexOffset != 0 ) {
    delete[] indexOffset;
    indexOffset = 0;
  }
  if ( activeTiles != 0 ) {
    delete[] activeTiles;
    activeTiles = 0;
  }

  if ( tileOffset != 0 ) {
    for ( int t = 0; t < TILINGS; ++t ) {
      delete[] tileOffset[t];
    }
    delete[] tileOffset;
    tileOffset = 0;
  }

  if ( weights != 0 ) {
    delete[] weights;
    weights = 0;
  }
  if ( eTraces != 0 ) {
    delete[] eTraces;
    eTraces = 0;
  }
}
