#ifndef __TILINGS_H__
#define __TILINGS_H__

#include "ValueFunction.h"
#include "Types.h"

#include <cstdio>
#include <list>
#include <map>
#include <set>

//class State;


class Tilings
{
  public:
    /**
    * Konstruktor
    * @param numberOfTilings - Anzahl der Tilings
    * @param varList - Liste mit den Daten der Variablen
    */
    Tilings( const int numberOfTilings, const std::list< Types::TILINGS_VAR_STRUCT >& varList );

    /**
    * Konstruktor zum Laden eines Tilings
    * @param aStatePtr - ein Pointer auf den Zustand
    * @param file - Filehandle
    */
    Tilings( State* aStatePtr, FILE* file );

    /**
    * Destruktor
    */
    virtual ~Tilings();

    /**
    * Initialisiert die Gewichte
    * @param initFunctionPtr - Pointer auf die Funktion die zum Initialisieren bentzt wird
    */
    virtual void initWeights( const double (*initFunctionPtr)() );

    /**
    * Initialisieren der EligibilityTraces mit 0.0
    */
    virtual void initEligibilityTraces();

    /**
    * Updaten der Gewichte
    * @param ALPHA - die Lernrate
    * @param DELTA - die Aenderung des Gewichts
    */
    virtual void updateWeights( const double ALPHA, const double DELTA );

    /**
    * Updaten der EligibilityTraces
    * @param GAMMA - Discountrate
    * @param LAMBDA - TraceDecay
    */
    virtual void updateEligibilityTraces( const double GAMMA, const double LAMBDA );

    /**
    * Ueberprueft ob die Gewichte in den Grenzen sind.
    * minWeight <= weight <= maxWeight
    * @param MIN_WEIGHT - minimales Gewicht
    * @param MAX_WEIGHT - maximales Gewicht
    */
    virtual void checkWeightLimit( const double MIN_WEIGHT, const double MAX_WEIGHT );

    /**
    * Liefert das minimale Gewicht zurueck.
    * @return das minimale Gewicht
    */
    virtual const double getMinWeight() const;

    /**
    * Liefert die Summe ueber alle Gewichte zurueck.
    * @param abs - wenn true wird der absolute Betrag der Gewichte aufsummiert
    * @param weightCount - Referenz auf die Anzahl der Gewichte (fuer alle Tilings)
    * @return das durchschnittliche Gewicht
    */
    virtual const double getWeightSum( const bool abs, int& weightCount ) const;

    /**
    * Liefert das maximale Gewicht zurueck.
    * @return das maximale Gewicht
    */
    virtual const double getMaxWeight() const;

    /**
    * Ermittelt die Indizes der aktivierten Features.
    */
    virtual void activateTiles();

    /**
    * Initialisiert die episodeStateHistory.
    */
    virtual void initEpisodeStateHistory();

    /**
    * Bewertet die gesamte Episode.
    * --> dafuer muss episodeStateHistory auf true gesetzt sein
    * @param value - dieser Wert wird auf alle Gewichte der Episode gerechnet
    */
    virtual void valueEpisode( const double value );

    /**
    * Abspeichern des Tilings
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int save( FILE* file ) const;

    /**
    * Laden der Tilingsdaten
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int load( FILE* file );

    /**
    * Liefert die Anzahl der Tilings.
    * @return Anzahl der Tilings
    */
    inline const int getNumberOfTilings() const;

    /**
    * Setzt die Art wie die Traces gehalten werden sollen.
    */
    inline void setTraceStorage( const Types::TRACE_STORAGE newTraceStorage );


  protected:
    /** Speicher fuer die Arrays holen. */
    virtual void allocateMemory();

    /**
    * Speicher fuer weights holen.
    * Da sich die Groesse des Arrays aus den anderen Werten errechnet, gibt
    * es 2 allocate Methoden.
    */
    virtual double* allocateMemoryWeights();

    /** Speicher freigeben */
    virtual void freeMemory();

    /**
    * Leert die TraceMap.
    */
    inline void clearTraceMap();


    /** Anzahl der Tilings (kann leider nicht const sein wegen laden) */
    int TILINGS;
    /** Anzahl der Dimensionen (kann leider nicht const sein wegen laden) */
    int DIMENSIONS;

    /** Wie werden die Traces gespeichert Array oder Map? */
    Types::TRACE_STORAGE traceStorage;

    /** Die Daten der Variablen */
    Types::TILINGS_VAR_STRUCT* vars;
    /** Anzahl der fuer jede Variable verwendeten Tiles */
    int* tiles;
    /** Groesse eines Tiles fuer jede Variable. */
    double* tileWidth;
    /**
    * Indexoffset der Variablen fuer weights und eTraces
    * --> damit der Index berechnet werden kann
    */
    int* indexOffset;

    /** Gibt fuer jedes Tiling das aktivierte Tile an */
    int* activeTiles;
    /** Speichert die aktivierten Tiles */
    int* storedActiveIndices;

    /** Fuer jedes Tiling und jede Variable das Offset */
    double** tileOffset;

    /**
    * Die Anzahl der Gewichte --> leider keine Konstante da der Wert erst
    * spaeter ermittelt wird, sollte aber als solche behandelt werden
    */
    int ARRAY_SIZE;
    /** Die zu lernenden Werte */
    double* weights;
    /** Die Eligibilitytraces --> Array */
    double* eTraces;
    /** Die Eligibilitytraces --> Map */
    std::map< int, double > eTracesMap;

    /** Pointer auf den Zustand */
    State* statePtr;

    /** Speichert alle Indizees dieser Episode */
    std::set< int > episodeIndexHistorySet;
};


inline const int Tilings::getNumberOfTilings() const
{
  return TILINGS;
}

inline void Tilings::setTraceStorage( const Types::TRACE_STORAGE newTraceStorage )
{
  traceStorage = newTraceStorage;
}

inline void Tilings::clearTraceMap()
{
  eTracesMap.clear();
}

#endif
