#ifndef __UTILITY_H__
#define __UTILITY_H__

#include <cmath>


namespace utility
{
  /**
  * verschiebt einen Wert an den Rand der Grenzen, sofern der Wert
  * ausserhalb dieser Grenzen lag
  * @param value der zu prfende Wert
  * @param lowerBound die untere Grenze
  * @param upperBound die obere Grenze
  * @return lowerBound  falls(value<lowerBound) ; upperBound  fallse(value>upperBound) ; value sonst
  */
  template< class T >
  inline T minMax( T value, T lowerBound, T upperBound )
  {
    value = ((value > lowerBound) ? (value) : (lowerBound));
    return ((value < upperBound) ? (value) : (upperBound));
  }

  /**
  * Normalisiert einen Wert.
  * @param value - der zu normalisierende Wert
  * @param minValue - der minimale Wert
  * @param maxValue - der maximale Wert
  * @return der normalisierte Wert
  */
  template< class T >
  inline T normalize( T value, T minValue, T maxValue )
  {
    T range = maxValue - minValue;
    while ( value < minValue ) {
      value += range;
    }
    while( value > maxValue ) {
      value -= range;
    }
    return value;
  }


  /**
  * Seed setzen.
  */
  inline void setSeed( unsigned int seed )
  {
    srand( seed );
  }

  /**
  * Liefert einen zufaelligen double-Wert aus [0,1)
  * @return double-Wert aus [0,1)
  */
  inline const double getNextDouble()
  {
    return static_cast< double >( rand() / (RAND_MAX + 1.0) );
  }

  /**
  * Waehlt aus einem Array von Werten einen Wert aus. Dazu wird eine Boltzmannverteilung
  * verwendet, die sich aus dem Array ergibt.
  * @param VALUES - Anzahl der Werte
  * @param valuePtr - Pointer auf ein Array mit Werten --> Die Werte werden veraendert !!!
  * @param TEMPERATURE - Temperatur-parameter der Boltzmannverteilung
  * @param MIN_VALUE - der kleinste Wert
  * @param MAX_VALUE - der groesste Wert
  * @return index des erwaehlten Wertes
  */
  inline const int getSoftMaxIndex( const int VALUES, double* valuePtr, const double TEMPERATURE,
                                    const double MIN_VALUE, const double MAX_VALUE )
  {
    double MAX_MIN_DIFF = MAX_VALUE - MIN_VALUE;
    if ( MAX_MIN_DIFF == 0.0 ) {
      // alle Werte sind gleich --> Gleichverteilung
      return minMax( static_cast< int >( floor(getNextDouble() * VALUES) ), 0, VALUES - 1 );
    }

    // Verteilung ermitteln
    double sum = 0.0;
    int index = 0;

    for ( index = 0; index < VALUES; ++index ) {
      // der Wert wird in das Intervall 0-1 gebracht --> eigentlich nicht so gut
      // aber bei grossen Werten laeuft man sonst Gefahr nan oder inf bei exp zu erhalten
      valuePtr[index] = std::exp( (valuePtr[index] - MIN_VALUE) / (MAX_MIN_DIFF * TEMPERATURE) );
      sum += valuePtr[index];
    }
    // hier werden 2 Schritte zusammengefasst
    // 1. die WK bestimmen (valuePtr / sum)
    // 2. die Auswahl des Wertes ueber die Summe
    const double DBL_RAND_MAX = static_cast< double >( RAND_MAX );
    const double RANDOM_VALUE = rand();
    double randSum = 0.0;
    for ( index = 0; index < VALUES; ++index ) {
      randSum += DBL_RAND_MAX * (valuePtr[index] / sum);
      if ( randSum > RANDOM_VALUE ) {
        break;
      }
    }

    return index;
  }
}

#endif
