#ifndef __VALUE_FUNCTION_H__
#define __VALUE_FUNCTION_H__

#include <string>

class State;


class ValueFunction
{
  public:
    /**
    * Konstruktor
    * @param aStatePtr - Pointer auf den Zustand
    */
    ValueFunction( State* aStatePtr );

    /**
    * Destruktor
    */
    virtual ~ValueFunction();

    /**
    * Initialisiert die Gewichte
    * @param initFunctionPtr - Pointer auf die Funktion die zum Initialisieren bentzt wird
    */
    virtual void initWeights( const double (*initFunctionPtr)() ) = 0;

    /**
    * Initialisieren der EligibilityTraces mit 0.0
    */
    virtual void initEligibilityTraces() = 0;

    /**
    * Updaten der Gewichte
    * @param ALPHA - die Lernrate
    * @param DELTA - die Aenderung des Gewichts
    */
    virtual void updateWeights( const double ALPHA, const double DELTA ) = 0;

    /**
    * Updaten der EligibilityTraces
    * @param GAMMA - Discountrate
    * @param LAMBDA - TraceDecay
    */
    virtual void updateEligibilityTraces( const double GAMMA, const double LAMBDA ) = 0;

    /**
    * Die Traces fuer die aktiven Indizees werden erhoeht.
    * @param VALUE - der Wert um den erhoeht wird
    */
    virtual void accumulatingTraces( const double VALUE ) = 0;

    /**
    * Die Traces fuer die aktiven Indizees werden gesetzt.
    * @param VALUE - der Wert auf den gesetzt wird
    */
    virtual void replacingTraces( const double VALUE ) = 0;

    /**
    * Ueberprueft ob die Gewichte in den Grenzen sind.
    * minWeight <= weight <= maxWeight
    */
    virtual void checkWeightLimit() = 0;

    /**
    * Liefert das minimale Gewicht zurueck.
    * @return das minimale Gewicht
    */
    virtual const double getMinWeight() const = 0;

    /**
    * Liefert das durchschnittliche Gewicht zurueck.
    * @param abs - wenn true wird der Durchschnitt ueber den absoluten Betrag der Gewichte gebildet
    * @return das durchschnittliche Gewicht
    */
    virtual const double getAvgWeight( const bool abs ) const = 0;

    /**
    * Liefert das maximale Gewicht zurueck.
    * @return das maximale Gewicht
    */
    virtual const double getMaxWeight() const = 0;

    /**
    * Setzt den Zustand
    */
    virtual void setState() = 0;

    /**
    * Falls keine EligibilityTraces verwendet werden, muss der Zustand gespeichert
    * werden, damit man weiss welche Gewichte geupdatet werden muessen.
    */
    virtual void storeState() = 0;

    /**
    * Initialisiert die episodeStateHistory.
    */
    virtual void initEpisodeStateHistory() = 0;

    /**
    * Bewertet die gesamte Episode.
    * --> dafuer muss episodeStateHistory auf true gesetzt sein
    * @param value - dieser Wert wird auf alle Gewichte der Episode gerechnet
    */
    virtual void valueEpisode( const double value ) = 0;

    /**
    * Abspeichern
    * @param fileName - Name der Datei die geschrieben werden soll
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int save( const std::string& fileName ) const = 0;

    /**
    * Laden
    * @param fileName - Name der Datei die geladen werden soll
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int load( const std::string& fileName ) = 0;

    /**
    * Liefert den StatePointer.
    */
    inline State* getStatePtr() const;

    /**
    * Setzt den minimalen Wert den ein Gewicht haben darf.
    * @param limit - das neue Minimum
    */
    inline void setMinWeightLimit( const double limit );

    /**
    * Setzt den maximalen Wert den ein Gewicht haben darf.
    * @param limit - das neue Maximum
    */
    inline void setMaxWeightLimit( const double limit );

    /**
    * Setzt ob eine Historie ueber die Zustaende gehalten werden soll.
    * Wird benoetigt, wenn man bspw. Episoden verstaerken/abschwaechen will.
    * @param flag - der neue Wert von keepStateHistory
    */
    inline void setEpisodeStateHistory( const bool flag );

    /**
    * Liefert ob eine Historie ueber die Zustaende gehalten werden soll
    * @return episodeStateHistory
    */
    inline const bool getEpisodeStateHistory() const;

    /**
    * Setzt value auf einen Wert.
    * @param aValue - der Wert
    */
    static inline void setValue( const double aValue );

    /**
    * Setzt die Grenzen fuer den zufaelligen Wert von getRandomValue.
    * @param aMinValue - untere Grenze
    * @param aMaxValue - obere Grenze
    */
    static inline void setRandomValueBoundaries( const double aMinValue, const double aMaxValue );

    /**
    * Setzt den Funktionspointer auf die Funktion mit der die Gewichte initialisiert werden sollen.
    */
    inline void setInitWeightsFunctionPtr( const double (*aFunctionPtr)() );


  protected:
    /**
    * Abspeichern
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int save( FILE* file ) const = 0;

    /**
    * Laden
    * @param file - Filehandle
    * @return 0 bei Erfolg, -1 bei Fehler
    */
    virtual const int load( FILE* file ) = 0;

    /**
    * Liefert einen festen, zuvor festgelegten double.
    * @return value
    */
    static inline const double getValue();

    /**
    * Liefert einen gleichverteilten zufaelligen Wert aus dem Intervall [minValue, maxValue).
    * @return gleichverteilter zufaelliger Wert
    */
    static inline const double getRandomValue();


    /** Pointer auf den Zustand. */
    State* statePtr;

    /** Pointer auf die Funktion die bei InitWeights verwendet wird. */
    const double (*initWeightsFunctionPtr)();

    /** Dieser Wert wird von getValue zurueckgegeben. */
    static double value;

    /** Die untere Grenze fuer getRandomValue. */
    static double minValue;

    /** Die obere Grenze fuer getRandomValue. */
    static double maxValue;

    /** Minimaler Wert eines Gewichts */
    double minWeightLimit;

    /** Maximaler Wert eines Gewichts */
    double maxWeightLimit;

    /** Soll eine Historie ueber die Zustaende einer Episode gehalten werden? */
    bool episodeStateHistory;
};

inline State* ValueFunction::getStatePtr() const
{
  return statePtr;
}

inline void ValueFunction::setMinWeightLimit( const double limit )
{
  minWeightLimit = limit;
}

inline void ValueFunction::setMaxWeightLimit( const double limit )
{
  maxWeightLimit = limit;
}

inline void ValueFunction::setEpisodeStateHistory( const bool flag )
{
  episodeStateHistory = flag;
}

inline const bool ValueFunction::getEpisodeStateHistory() const
{
  return episodeStateHistory;
}

inline void ValueFunction::setValue( const double aValue )
{
  value = aValue;
}

inline void ValueFunction::setRandomValueBoundaries( const double aMinValue, const double aMaxValue )
{
  minValue = aMinValue;
  maxValue = aMaxValue;
}

inline void ValueFunction::setInitWeightsFunctionPtr( const double (*aFunctionPtr)() )
{
  initWeightsFunctionPtr = aFunctionPtr;
}

inline const double ValueFunction::getValue()
{
  return value;
}

inline const double ValueFunction::getRandomValue()
{
  return (minValue + static_cast< double >( rand() / (RAND_MAX + 1.0) ) * (maxValue - minValue));
}

#endif
