#include <AbstractAction.h>
#include <ActionLookupTable.h>
#include <Log.h>
#include <Sarsa.h>
#include <State.h>

class Action : public AbstractAction
{
  public:
    Action( const int ew, const int ns )
      : AbstractAction()
    {
      eastWest = ew;
      northSouth = ns;
    }

    Action( FILE* file )
    {
      load( file );
    }

    virtual void save( FILE* file ) const
    {
      fwrite( &eastWest, sizeof( int ), 1, file );
      fwrite( &northSouth, sizeof( int ), 1, file );
    }

    virtual void load( FILE* file )
    {
      fread( &eastWest, sizeof( int ), 1, file );
      fread( &northSouth, sizeof( int ), 1, file );
    }

    inline const int getEastWest() const
    {
      return eastWest;
    }

    inline const int getNorthSouth() const
    {
      return northSouth;
    }


  private:
    int eastWest;
    int northSouth;
};


class GridWorldAction : public Sarsa
{
  public:
    GridWorldAction( const std::string confFile, const std::string valueFunctionFile,
                     const std::string actionFile, const bool loadValueFunction )
      : Sarsa( confFile ),
        MIN_COL(1),
        MAX_COL(10),
        MIN_ROW(1),
        MAX_ROW(7),
        START_COL(1),
        START_ROW(4),
        GOAL_COL(8),
        GOAL_ROW(4),
        VALUE_FUNCTION_FILE( valueFunctionFile ),
        ACTION_FILE( actionFile )
    {
      // init windarray
      wind = new int[ MAX_COL - MIN_COL + 1 ];
      wind[0] = 0;
      wind[1] = 0;
      wind[2] = 0;
      wind[3] = 1;
      wind[4] = 1;
      wind[5] = 1;
      wind[6] = 2;
      wind[7] = 2;
      wind[8] = 1;
      wind[9] = 0;

      if ( loadValueFunction ) {
        // load all objects
        load();
      }
      else {
        // create all objects
        // create Stateobject
        State* statePtr = new State();
        colId = statePtr->registerVariable( "COL", MIN_COL, MAX_COL );
        rowId = statePtr->registerVariable( "ROW", MIN_ROW, MAX_ROW );

        // create ValueFunction
        avfPtr = new ActionLookupTable( statePtr );

        // create and register Actions
        // Go East
        avfPtr->registerAction( new Action(  1,  0 ) );
        // Go West
        avfPtr->registerAction( new Action( -1,  0 ) );
        // Go North
        avfPtr->registerAction( new Action(  0,  1 ) );
        // Go South
        avfPtr->registerAction( new Action(  0, -1 ) );

        static_cast< ActionLookupTable* >( avfPtr )->create();
      }

      // statistics
      stepSum = 0;
    }


    ~GridWorldAction()
    {
      // nothing to do
    }


  protected:
    virtual void setInitialState()
    {
      currentCol = START_COL;
      currentRow = START_ROW;
      avfPtr->getStatePtr()->setValueInt( colId, currentCol );
      avfPtr->getStatePtr()->setValueInt( rowId, currentRow );
    }


    virtual void takeAction()
    {
      const Action* action = static_cast< const Action* >( avfPtr->getActiveAction() );

      // change position
      currentRow += action->getNorthSouth() + wind[ currentCol - 1 ];
      currentCol += action->getEastWest();

      // check world-limits
      if ( currentCol < MIN_COL ) {
        currentCol = MIN_COL;
      }
      if ( currentCol > MAX_COL ) {
        currentCol = MAX_COL;
      }
      if ( currentRow < MIN_ROW ) {
        currentRow = MIN_ROW;
      }
      if ( currentRow > MAX_ROW ) {
        currentRow = MAX_ROW;
      }
    }


    virtual const bool observeState()
    {
      avfPtr->getStatePtr()->setValueInt( colId, currentCol );
      avfPtr->getStatePtr()->setValueInt( rowId, currentRow );

      if ( (currentCol == GOAL_COL) && (currentRow == GOAL_ROW) ) {
        return false;
      }
      return true;
    }


    virtual const double calculateReward()
    {
      if ( (currentCol == GOAL_COL) && (currentRow == GOAL_ROW) ) {
        return 1.0;
      }
      else {
        return -1.0;
      }
    }


    virtual void finishEpisode()
    {
      stepSum += steps;
      std::cout << episodes << "  Steps: " << steps << "   Average: " << (stepSum / episodes) << std::endl;

      // save valuefunction and actions
      save();
    }

    virtual void save() const
    {
      // save valuefunction
      avfPtr->save( VALUE_FUNCTION_FILE );

      // open file
      FILE *file = fopen( ACTION_FILE.c_str(), "w+" );
      // save actions
      for ( int a = 0; a < avfPtr->getNumberOfActions(); ++a ) {
        static_cast< const Action* >( avfPtr->getAction( a ) )->save( file );
      }
      // close file
      fclose( file );
    }

    virtual void load()
    {
      // load valuefunction
      avfPtr = new ActionLookupTable( VALUE_FUNCTION_FILE );

      // open file
      FILE *file = fopen( ACTION_FILE.c_str(), "r" );
      // load actions
      for ( int a = 0; a < avfPtr->getNumberOfActions(); ++a ) {
        avfPtr->setAction( a, new Action( file ) );
      }
      // close file
      fclose( file );

      // get var ids
      colId = avfPtr->getStatePtr()->getVariableId( "COL" );
      rowId = avfPtr->getStatePtr()->getVariableId( "ROW" );
    }


    // StateVariableIds
    int rowId;
    int colId;

    // the GridWorld-Boundaries
    const int MIN_COL;
    const int MAX_COL;
    const int MIN_ROW;
    const int MAX_ROW;

    // Start- and Goalposition
    const int START_COL;
    const int START_ROW;
    const int GOAL_COL;
    const int GOAL_ROW;

    // wind in the gridworld --> affects only rows
    int* wind;

    // current position in the world
    int currentCol;
    int currentRow;

    double stepSum;

    // files
    const std::string VALUE_FUNCTION_FILE;
    const std::string ACTION_FILE;
};



int main( int argC, char** argV )
{
  std::string confFile = "gridworldaction.conf";
  std::string valueFunctionFile = "gridworldaction.valuefunction";
  std::string actionFile = "gridworldaction.action";
  bool loadValueFunction = false;

  bool error = false;
  for ( int i = 1; i < argC; ++i ) {
    std::string param = std::string( argV[i] );
    if ( param == "-conf" ) {
      if ( ++i < argC ) {
        confFile = std::string( argV[i] );
      }
      else {
        error = true;
        break;
      }
    }
    else if ( param == "-vf" ) {
      if ( ++i < argC ) {
        valueFunctionFile = std::string( argV[i] );
      }
      else {
        error = true;
        break;
      }
    }
    else if ( param == "-action" ) {
      if ( ++i < argC ) {
        actionFile = std::string( argV[i] );
      }
      else {
        error = true;
        break;
      }
    }
    else if ( param == "-load" ) {
      loadValueFunction = true;
    }
    else {
      error = true;
      break;
    }
  }

  if ( error == true ) {
    std::cerr << "Usage: gwaction [-conf <configuration>] [-vf <valuefunction>] [-action <actions>] [-load]" << std::endl;
    return -1;
  }

  GridWorldAction gridworld = GridWorldAction( confFile, valueFunctionFile, actionFile, loadValueFunction );

  // stop with CTRL-C
  gridworld.controlLoop();

  return 0;
}
