#include <StateLookupTable.h>
#include <Log.h>
#include <TD.h>
#include <State.h>
#include <Utility.h>

#include <map>

class Action
{
  public:
    Action( const int ew, const int ns )
    {
      eastWest = ew;
      northSouth = ns;
    }

    inline const int getEastWest() const
    {
      return eastWest;
    }


    inline const int getNorthSouth() const
    {
      return northSouth;
    }


  private:
    int eastWest;
    int northSouth;
};


class GridWorldState : public TD
{
  public:
    GridWorldState( const std::string confFile, const std::string valueFunctionFile,
                    const bool loadValueFunction )
      : TD( confFile ),
        MIN_COL(1),
        MAX_COL(10),
        MIN_ROW(1),
        MAX_ROW(7),
        START_COL(1),
        START_ROW(4),
        GOAL_COL(8),
        GOAL_ROW(4),
        VALUE_FUNCTION_FILE( valueFunctionFile )
    {
      // init windarray
      wind = new int[ MAX_COL - MIN_COL + 1 ];
      wind[0] = 0;
      wind[1] = 0;
      wind[2] = 0;
      wind[3] = 1;
      wind[4] = 1;
      wind[5] = 1;
      wind[6] = 2;
      wind[7] = 2;
      wind[8] = 1;
      wind[9] = 0;

      // create Actions
      // Go East
      actions[0] = new Action(  1,  0 );
      // Go West
      actions[1] = new Action( -1,  0 );
      // Go North
      actions[2] = new Action(  0,  1 );
      // Go South
      actions[3] = new Action(  0, -1 );

      if ( loadValueFunction ) {
        // load valuefunction
        svfPtr = new StateLookupTable( VALUE_FUNCTION_FILE );

        // get var ids
        colId = svfPtr->getStatePtr()->getVariableId( "COL" );
        rowId = svfPtr->getStatePtr()->getVariableId( "ROW" );
      }
      else {
        // create Stateobject
        State* statePtr = new State();
        colId = statePtr->registerVariable( "COL", MIN_COL, MAX_COL );
        rowId = statePtr->registerVariable( "ROW", MIN_ROW, MAX_ROW );

        // create ValueFunction
        svfPtr = new StateLookupTable( statePtr );
      }

      // statistics
      stepSum = 0;
    }


    ~GridWorldState()
    {
      // nothing to do
    }


  protected:
    virtual void setInitialState()
    {
      currentCol = START_COL;
      currentRow = START_ROW;
      svfPtr->getStatePtr()->setValueInt( colId, currentCol );
      svfPtr->getStatePtr()->setValueInt( rowId, currentRow );
    }


    virtual void takeAction()
    {
      int index = 0;

      if ( randomAction() == false ) {
        // Greedy
        // evaluate all states
        std::map< int, double > indexStateValueMap;
        for ( int a = 0; a < 4; ++a ) {
          // get possible state
          int succRow = currentRow + actions[a]->getNorthSouth() + wind[ currentCol - 1 ];
          int succCol = currentCol + actions[a]->getEastWest();

          // set state
          svfPtr->getStatePtr()->setValueInt( colId, succCol );
          svfPtr->getStatePtr()->setValueInt( rowId, succRow );
          svfPtr->setState();

          // get statevalue
          double stateValue = svfPtr->getStateValue();

          if ( indexStateValueMap.empty() ) {
            indexStateValueMap[ a ] = stateValue;
          }
          else if ( stateValue == indexStateValueMap.begin()->second ) {
            indexStateValueMap[ a ] = stateValue;
          }
          else if ( stateValue > indexStateValueMap.begin()->second ) {
            indexStateValueMap.clear();
            indexStateValueMap[ a ] = stateValue;
          }
        }

        // choose best state
        int element = static_cast< int >( utility::getNextDouble() * indexStateValueMap.size() );
        for ( std::map< int, double >::const_iterator it = indexStateValueMap.begin();
              element >= 0; (--element, ++it) )
        {
          index = it->first;
        }
      }
      else {
        // Random
        index = static_cast< int >( utility::getNextDouble() * 4 );
      }

      // change position
      currentRow += actions[index]->getNorthSouth() + wind[ currentCol - 1 ];
      currentCol += actions[index]->getEastWest();

      // check world-limits
      if ( currentCol < MIN_COL ) {
        currentCol = MIN_COL;
      }
      if ( currentCol > MAX_COL ) {
        currentCol = MAX_COL;
      }
      if ( currentRow < MIN_ROW ) {
        currentRow = MIN_ROW;
      }
      if ( currentRow > MAX_ROW ) {
        currentRow = MAX_ROW;
      }
    }


    virtual const bool observeState()
    {
      svfPtr->getStatePtr()->setValueInt( colId, currentCol );
      svfPtr->getStatePtr()->setValueInt( rowId, currentRow );

      if ( (currentCol == GOAL_COL) && (currentRow == GOAL_ROW) ) {
        return false;
      }
      return true;
    }


    virtual const double calculateReward()
    {
      if ( (currentCol == GOAL_COL) && (currentRow == GOAL_ROW) ) {
        return 1.0;
      }
      else {
        return -1.0;
      }
    }


    virtual void finishEpisode()
    {
      stepSum += steps;
      std::cout << episodes << "  Steps: " << steps << "   Average: " << (stepSum / episodes) << std::endl;

      // save valuefunction
      svfPtr->save( VALUE_FUNCTION_FILE );
    }


    // StateVariableIds
    int rowId;
    int colId;

    // the GridWorld-Boundaries
    const int MIN_COL;
    const int MAX_COL;
    const int MIN_ROW;
    const int MAX_ROW;

    // Start- and Goalposition
    const int START_COL;
    const int START_ROW;
    const int GOAL_COL;
    const int GOAL_ROW;

    // wind in the gridworld --> affects only rows
    int* wind;

    // current position in the world
    int currentCol;
    int currentRow;

    Action* actions[4];

    double stepSum;

    // file
    const std::string VALUE_FUNCTION_FILE;
};



int main( int argC, char** argV )
{
  std::string confFile = "gridworldstate.conf";
  std::string valueFunctionFile = "gridworldstate.valuefunction";
  bool loadValueFunction = false;

  bool error = false;
  for ( int i = 1; i < argC; ++i ) {
    std::string param = std::string( argV[i] );
    if ( param == "-conf" ) {
      if ( ++i < argC ) {
        confFile = std::string( argV[i] );
      }
      else {
        error = true;
        break;
      }
    }
    else if ( param == "-vf" ) {
      if ( ++i < argC ) {
        valueFunctionFile = std::string( argV[i] );
      }
      else {
        error = true;
        break;
      }
    }
    else if ( param == "-load" ) {
      loadValueFunction = true;
    }
    else {
      error = true;
      break;
    }
  }

  if ( error == true ) {
    std::cerr << "Usage: gwstate [-conf <configuration>] [-vf <valuefunction>] [-load]" << std::endl;
    return -1;
  }

  GridWorldState gridworld = GridWorldState( confFile, valueFunctionFile, loadValueFunction );

  // stop with CTRL-C
  gridworld.controlLoop();

  return 0;
}
