#include <AbstractAction.h>
#include <ActionTileCoding.h>
#include <ActionTilings.h>
#include <Log.h>
#include <Sarsa.h>
#include <State.h>
#include <Types.h>
#include <Utility.h>

#include <list>

class Action : public AbstractAction
{
  public:
    Action( double acc )
      : AbstractAction()
    {
      acceleration = acc;
    }

    Action( FILE* file )
    {
      load( file );
    }

    virtual void save( FILE* file ) const
    {
      fwrite( &acceleration, sizeof( double ), 1, file );
    }

    virtual void load( FILE* file )
    {
      fread( &acceleration, sizeof( double ), 1, file );
    }

    inline const double getAcceleration() const
    {
      return acceleration;
    }


  private:
    double acceleration;
};


class MountainCarAction : public Sarsa
{
  public:
    MountainCarAction( const std::string confFile, const std::string valueFunctionFile,
                       const std::string actionFile, const bool loadValueFunction )
      : Sarsa( confFile ),
        MIN_POS(-1.2),
        MAX_POS( 0.5),
        MIN_VEL(-0.07),
        MAX_VEL( 0.07),
        VALUE_FUNCTION_FILE( valueFunctionFile ),
        ACTION_FILE( actionFile )
    {
      if ( loadValueFunction ) {
        // load all objects
        load();
      }
      else {
        // create all objects
        // create Stateobject
        State* statePtr = new State();
        posId = statePtr->registerVariable( "POS", MIN_POS, MAX_POS );
        velId = statePtr->registerVariable( "VEL", MIN_VEL, MAX_VEL );

        // create ValueFunction
        avfPtr = new ActionTileCoding( statePtr );

        // create and register Actions
        // Backward
        avfPtr->registerAction( new Action( -1.0 ) );
        // No Acceleration
        avfPtr->registerAction( new Action(  0.0 ) );
        // Forward
        avfPtr->registerAction( new Action(  1.0 ) );

        // die Tilings
        Types::TILINGS_VAR_STRUCT vs;
        std::list< Types::TILINGS_VAR_STRUCT > vsList;

        // Tilings 1
        vs.id = posId;
        vs.resolution = 0.018;
        vs.tileType = Types::WRAP;
        vsList.push_back( vs );
        vs.id = velId;
        vs.resolution = 0.0015;
        vs.tileType = Types::WRAP;
        vsList.push_back( vs );

        ActionTilings* t = new ActionTilings( 4, vsList );
        static_cast< ActionTileCoding* >(avfPtr)->registerTiling( t );
      }

      // statistics
      stepSum = 0;
    }


    ~MountainCarAction()
    {
      // nothing to do
    }


  protected:
    virtual void setInitialState()
    {
      currentPos = MIN_POS + utility::getNextDouble() * (MAX_POS - MIN_POS);
      currentVel = MIN_VEL + utility::getNextDouble() * (MAX_VEL - MIN_VEL);
      avfPtr->getStatePtr()->setValueDbl( posId, currentPos );
      avfPtr->getStatePtr()->setValueDbl( velId, currentVel );
    }


    virtual void takeAction()
    {
      const Action* action = static_cast< const Action* >( avfPtr->getActiveAction() );

      // change velocity
      currentVel += 0.001 * action->getAcceleration() - 0.0025 * cos( 3.0 * currentPos );
      if ( currentVel < MIN_VEL ) {
        currentVel = MIN_VEL;
      }
      if ( currentVel > MAX_VEL ) {
        currentVel = MAX_VEL;
      }

      // change position
      currentPos += currentVel;
      if ( currentPos < MIN_POS ) {
        currentPos = MIN_POS;
        currentVel = 0.0;
      }
      if ( currentPos >= MAX_POS ) {
        // do nothing --> we're done
      }
    }


    virtual const bool observeState()
    {
      avfPtr->getStatePtr()->setValueDbl( posId, currentPos );
      avfPtr->getStatePtr()->setValueDbl( velId, currentVel );

      if ( currentPos >= MAX_POS ) {
        return false;
      }
      return true;
    }


    virtual const double calculateReward()
    {
      if ( currentPos >= MAX_POS ) {
        return 1.0;
      }
      else {
        return -1.0;
      }
    }


    virtual void finishEpisode()
    {
      stepSum += steps;
      std::cout << episodes << "  Steps: " << steps << "   Average: " << (stepSum / episodes) << std::endl;

      // save valuefunction and actions
      save();
    }

    virtual void save() const
    {
      // save valuefunction
      avfPtr->save( VALUE_FUNCTION_FILE );

      // open file
      FILE *file = fopen( ACTION_FILE.c_str(), "w+" );
      // save actions
      for ( int a = 0; a < avfPtr->getNumberOfActions(); ++a ) {
        static_cast< const Action* >( avfPtr->getAction( a ) )->save( file );
      }
      // close file
      fclose( file );
    }

    virtual void load()
    {
      // load valuefunction
      avfPtr = new ActionTileCoding( VALUE_FUNCTION_FILE );

      // open file
      FILE *file = fopen( ACTION_FILE.c_str(), "r" );
      // load actions
      for ( int a = 0; a < avfPtr->getNumberOfActions(); ++a ) {
        avfPtr->setAction( a, new Action( file ) );
      }
      // close file
      fclose( file );

      // get var ids
      posId = avfPtr->getStatePtr()->getVariableId( "POS" );
      velId = avfPtr->getStatePtr()->getVariableId( "VEL" );
    }


    // StateVariableIds
    int posId;
    int velId;

    // the World-Boundaries
    const double MIN_POS;
    const double MAX_POS;
    const double MIN_VEL;
    const double MAX_VEL;

    // current position in the world
    double currentPos;
    double currentVel;

    double stepSum;

    // files
    const std::string VALUE_FUNCTION_FILE;
    const std::string ACTION_FILE;
};



int main( int argC, char** argV )
{
  std::string confFile = "mountaincaraction.conf";
  std::string valueFunctionFile = "mountaincaraction.valuefunction";
  std::string actionFile = "mountaincaraction.action";
  bool loadValueFunction = false;

  bool error = false;
  for ( int i = 1; i < argC; ++i ) {
    std::string param = std::string( argV[i] );
    if ( param == "-conf" ) {
      if ( ++i < argC ) {
        confFile = std::string( argV[i] );
      }
      else {
        error = true;
        break;
      }
    }
    else if ( param == "-vf" ) {
      if ( ++i < argC ) {
        valueFunctionFile = std::string( argV[i] );
      }
      else {
        error = true;
        break;
      }
    }
    else if ( param == "-action" ) {
      if ( ++i < argC ) {
        actionFile = std::string( argV[i] );
      }
      else {
        error = true;
        break;
      }
    }
    else if ( param == "-load" ) {
      loadValueFunction = true;
    }
    else {
      error = true;
      break;
    }
  }

  if ( error == true ) {
    std::cerr << "Usage: mcaction [-conf <configuration>] [-vf <valuefunction>] [-action <actions>] [-load]" << std::endl;
    return -1;
  }

  MountainCarAction mountaincar = MountainCarAction( confFile, valueFunctionFile, actionFile, loadValueFunction );

  // stop with CTRL-C
  mountaincar.controlLoop();

  return 0;
}
