#include <Log.h>
#include <State.h>
#include <StateNeuralNetwork.h>
#include <TD.h>
#include <Utility.h>

#include <list>
#include <map>


class Action
{
  public:
    Action( double acc )
    {
      acceleration = acc;
    }

    inline const double getAcceleration() const
    {
      return acceleration;
    }


  private:
    double acceleration;
};


class MountainCarStateNet : public TD
{
  public:
    MountainCarStateNet( const std::string confFile, const std::string valueFunctionFile,
                         const bool loadValueFunction )
      : TD( confFile ),
        MIN_POS(-1.2),
        MAX_POS( 0.5),
        MIN_VEL(-0.07),
        MAX_VEL( 0.07),
        VALUE_FUNCTION_FILE( valueFunctionFile )
    {
      // create Actions
      actions[0] = new Action( -1.0 );
      actions[1] = new Action(  0.0 );
      actions[2] = new Action(  1.0 );

      if ( loadValueFunction ) {
        // load valuefunction
        svfPtr = new StateNeuralNetwork( VALUE_FUNCTION_FILE );

        // get var ids
        posId = svfPtr->getStatePtr()->getVariableId( "POS" );
        velId = svfPtr->getStatePtr()->getVariableId( "VEL" );
      }
      else {
        // create Stateobject
        State* statePtr = new State();
        posId = statePtr->registerVariable( "POS", MIN_POS, MAX_POS );
        velId = statePtr->registerVariable( "VEL", MIN_VEL, MAX_VEL );

        // create ValueFunction
        svfPtr = new StateNeuralNetwork( statePtr, 25 );
      }
      
      svfPtr->setMinWeightLimit(-5.0);
      svfPtr->setMaxWeightLimit( 5.0);
      
      // statistics
      stepSum = 0;
    }


    ~MountainCarStateNet()
    {
      // nothing to do
    }


  protected:
    virtual void setInitialState()
    {
      currentPos = MIN_POS + utility::getNextDouble() * (MAX_POS - MIN_POS);
      currentVel = MIN_VEL + utility::getNextDouble() * (MAX_VEL - MIN_VEL);
      svfPtr->getStatePtr()->setValueDbl( posId, currentPos );
      svfPtr->getStatePtr()->setValueDbl( velId, currentVel );
    }


    virtual void takeAction()
    {
      int index = 0;

      if ( randomAction() == false ) {
        // Greedy
        // evaluate all states
        std::map< int, double > indexStateValueMap;
        for ( int a = 0; a < 3; ++a ) {
          // get possible state
          double succVel = currentVel + 0.001 * actions[a]->getAcceleration() - 0.0025 * cos( 3.0 * currentPos );
          if ( succVel < MIN_VEL ) {
            succVel = MIN_VEL;
          }
          if ( succVel > MAX_VEL ) {
            succVel = MAX_VEL;
          }
          double succPos = currentPos + currentVel;
          if ( succPos < MIN_POS ) {
            succPos = MIN_POS;
            succVel = 0.0;
          }
          if ( succPos >= MAX_POS ) {
            // do nothing --> we're done
          }

          // set state
          svfPtr->getStatePtr()->setValueDbl( posId, succPos );
          svfPtr->getStatePtr()->setValueDbl( velId, succVel );
          svfPtr->setState();

          // get statevalue
          double stateValue = svfPtr->getStateValue();

          if ( indexStateValueMap.empty() ) {
            indexStateValueMap[ a ] = stateValue;
          }
          else if ( stateValue == indexStateValueMap.begin()->second ) {
            indexStateValueMap[ a ] = stateValue;
          }
          else if ( stateValue > indexStateValueMap.begin()->second ) {
            indexStateValueMap.clear();
            indexStateValueMap[ a ] = stateValue;
          }
        }

        // choose best state
        int element = static_cast< int >( utility::getNextDouble() * indexStateValueMap.size() );
        for ( std::map< int, double >::const_iterator it = indexStateValueMap.begin();
              element >= 0; (--element, ++it) )
        {
          index = it->first;
        }
      }
      else {
        // Random
        index = static_cast< int >( utility::getNextDouble() * 3 );
      }

      // change velocity
      currentVel += 0.001 * actions[index]->getAcceleration() - 0.0025 * cos( 3.0 * currentPos );
      if ( currentVel < MIN_VEL ) {
        currentVel = MIN_VEL;
      }
      if ( currentVel > MAX_VEL ) {
        currentVel = MAX_VEL;
      }

      // change position
      currentPos += currentVel;
      if ( currentPos < MIN_POS ) {
        currentPos = MIN_POS;
        currentVel = 0.0;
      }
      if ( currentPos >= MAX_POS ) {
        // do nothing --> we're done
      }
    }


    virtual const bool observeState()
    {
      svfPtr->getStatePtr()->setValueDbl( posId, currentPos );
      svfPtr->getStatePtr()->setValueDbl( velId, currentVel );

      if ( (currentPos >= MAX_POS) || (steps == 200) ) {
        return false;
      }
      return true;
    }


    virtual const double calculateReward()
    {
      if ( currentPos >= MAX_POS ) {
        return 10.0;
      }
      else {
        return -1.0;
      }
    }


    virtual void finishEpisode()
    {
      stepSum += steps;
      std::cout << episodes << "  Steps: " << steps << "   Average: " << (stepSum / episodes) << std::endl;

      // save valuefunction
      svfPtr->save( VALUE_FUNCTION_FILE );
    }


    // StateVariableIds
    int posId;
    int velId;

    // the World-Boundaries
    const double MIN_POS;
    const double MAX_POS;
    const double MIN_VEL;
    const double MAX_VEL;

    // current position in the world
    double currentPos;
    double currentVel;

    Action* actions[3];

    double stepSum;

    // file
    const std::string VALUE_FUNCTION_FILE;
};



int main( int argC, char** argV )
{
  std::string confFile = "mountaincarstatenet.conf";
  std::string valueFunctionFile = "mountaincarstatenet.valuefunction";
  bool loadValueFunction = false;

  bool error = false;
  for ( int i = 1; i < argC; ++i ) {
    std::string param = std::string( argV[i] );
    if ( param == "-conf" ) {
      if ( ++i < argC ) {
        confFile = std::string( argV[i] );
      }
      else {
        error = true;
        break;
      }
    }
    else if ( param == "-vf" ) {
      if ( ++i < argC ) {
        valueFunctionFile = std::string( argV[i] );
      }
      else {
        error = true;
        break;
      }
    }
    else if ( param == "-load" ) {
      loadValueFunction = true;
    }
    else {
      error = true;
      break;
    }
  }

  if ( error == true ) {
    std::cerr << "Usage: mcstatenet [-conf <configuration>] [-vf <valuefunction>] [-load]" << std::endl;
    return -1;
  }

  MountainCarStateNet mountaincar = MountainCarStateNet( confFile, valueFunctionFile, loadValueFunction );

  // stop with CTRL-C
  mountaincar.controlLoop();

  return 0;
}
